# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
from typing import Dict, Optional, Type

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....typing import BandType
from ...cluster import ClusterAPI
from ..core import Subtask, SubtaskResult
from ..errors import SlotOccupiedAlready
from .processor import SubtaskProcessor, SubtaskProcessorActor

logger = logging.getLogger(__name__)


SubtaskRunnerRef = mo.ActorRefType["SubtaskRunnerActor"]


class SubtaskRunnerActor(mo.Actor):
    _session_id_to_processors: Dict[str, mo.ActorRefType[SubtaskProcessorActor]]
    _running_processor: Optional[mo.ActorRefType[SubtaskProcessorActor]]
    _last_processor: Optional[mo.ActorRefType[SubtaskProcessorActor]]

    @classmethod
    def gen_uid(cls, band_name: str, slot_id: int):
        return f"slot_{band_name}_{slot_id}_subtask_runner"

    def __init__(
        self, band: BandType, worker_address: str, subtask_processor_cls: Type = None
    ):
        self._band = band
        self._worker_address = worker_address
        self._subtask_processor_cls = self._get_subtask_processor_cls(
            subtask_processor_cls
        )

        self._cluster_api = None

        self._session_id_to_processors = dict()
        self._running_processor = None
        self._last_processor = None

    async def __post_create__(self):
        self._cluster_api = await ClusterAPI.create(address=self.address)

    async def __pre_destroy__(self):
        try:
            await asyncio.gather(
                *[
                    mo.destroy_actor(ref)
                    for ref in self._session_id_to_processors.values()
                ]
            )
        except mo.ActorNotExist:  # pragma: no cover
            # deleted, ignore
            pass

    @classmethod
    def _get_subtask_processor_cls(cls, subtask_processor_cls):
        if subtask_processor_cls is None:
            return SubtaskProcessor
        else:
            assert isinstance(subtask_processor_cls, str)
            module, class_name = subtask_processor_cls.rsplit(".", 1)
            return getattr(importlib.import_module(module), class_name)

    async def _run_subtask(self, subtask: Subtask):
        processor = await self._init_subtask_processor(subtask)
        self._subtask_info.processor = processor
        return await processor.run()

    @alru_cache(cache_exceptions=False)
    async def _get_supervisor_address(self, session_id: str):
        [address] = await self._cluster_api.get_supervisors_by_keys([session_id])
        return address

    async def run_subtask(self, subtask: Subtask):
        if self._running_processor is not None:  # pragma: no cover
            running_subtask_id = await self._running_processor.get_running_subtask_id()
            # current subtask is still running
            raise SlotOccupiedAlready(
                f"There is subtask(id: {running_subtask_id}) running in {self.uid} "
                f"at {self.address}, cannot run subtask {subtask.subtask_id}"
            )

        session_id = subtask.session_id
        supervisor_address = await self._get_supervisor_address(session_id)
        if session_id not in self._session_id_to_processors:
            try:
                self._session_id_to_processors[session_id] = await mo.create_actor(
                    SubtaskProcessorActor,
                    session_id,
                    self._band,
                    supervisor_address,
                    self._worker_address,
                    self._subtask_processor_cls,
                    uid=SubtaskProcessorActor.gen_uid(session_id),
                    address=self.address,
                )
            except mo.ActorAlreadyExist:
                # when recovering actor pools, the actor created in sub pools
                # may be recovered already
                self._session_id_to_processors[session_id] = await mo.actor_ref(
                    uid=SubtaskProcessorActor.gen_uid(session_id),
                    address=self.address,
                )
        processor = self._session_id_to_processors[session_id]
        try:
            self._running_processor = self._last_processor = processor
            result = yield self._running_processor.run(subtask)
        finally:
            self._running_processor = None
        raise mo.Return(result)

    async def get_subtask_result(self) -> SubtaskResult:
        return self._last_processor.result()

    def is_runner_free(self):
        return self._running_processor is None

    async def cancel_subtask(self):
        if self._running_processor is None:
            return
        running_subtask_id = await self._running_processor.get_running_subtask_id()
        logger.info("Start to cancel subtask %s.", running_subtask_id)
        await self._running_processor.cancel()
        self._running_processor = None
        logger.info("Canceled subtask %s.", running_subtask_id)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType, recursive_tile
from ...config import options
from ...serialization.serializables import AnyField, BoolField, StringField, Int32Field
from ..align import align_dataframe_series
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, validate_axis


class DataFrameDropNA(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.DROP_NA

    _axis = AnyField("axis")
    _how = StringField("how")
    _thresh = Int32Field("thresh")
    _subset = AnyField("subset")
    _use_inf_as_na = BoolField("use_inf_as_na")

    # when True, dropna will be called on the input,
    # otherwise non-nan counts will be used
    _drop_directly = BoolField("drop_directly")
    # size of subset, used when how == 'any'
    _subset_size = Int32Field("subset_size")

    def __init__(
        self,
        axis=None,
        how=None,
        thresh=None,
        subset=None,
        use_inf_as_na=None,
        drop_directly=None,
        subset_size=None,
        sparse=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _axis=axis,
            _how=how,
            _thresh=thresh,
            _subset=subset,
            _use_inf_as_na=use_inf_as_na,
            _drop_directly=drop_directly,
            _subset_size=subset_size,
            _output_types=output_types,
            sparse=sparse,
            **kw
        )

    @property
    def axis(self) -> int:
        return self._axis

    @property
    def how(self) -> str:
        return self._how

    @property
    def thresh(self) -> int:
        return self._thresh

    @property
    def subset(self) -> list:
        return self._subset

    @property
    def use_inf_as_na(self) -> bool:
        return self._use_inf_as_na

    @property
    def drop_directly(self) -> bool:
        return self._drop_directly

    @property
    def subset_size(self) -> int:
        return self._subset_size

    def __call__(self, df):
        new_shape = list(df.shape)
        new_shape[0] = np.nan

        params = df.params.copy()
        params["index_value"] = parse_index(None, df.key, df.index_value.key)
        params["shape"] = tuple(new_shape)
        return self.new_tileable([df], **params)

    @classmethod
    def _tile_drop_directly(cls, op: "DataFrameDropNA"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_df.chunks:
            new_shape = list(c.shape)
            new_shape[0] = np.nan

            params = c.params.copy()
            params["index_value"] = parse_index(None, c.key, c.index_value.key)
            params["shape"] = tuple(new_shape)

            new_op = op.copy().reset_key()
            new_op._drop_directly = True
            chunks.append(new_op.new_chunk([c], **params))

        new_nsplits = list(in_df.nsplits)
        new_nsplits[0] = (np.nan,) * len(in_df.nsplits[0])

        new_op = op.copy().reset_key()
        params = out_df.params.copy()
        params.update(dict(chunks=chunks, nsplits=new_nsplits))
        return new_op.new_tileables([in_df], **params)

    @classmethod
    def tile(cls, op: "DataFrameDropNA"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        # series tiling will go here
        if len(in_df.chunk_shape) == 1 or in_df.chunk_shape[1] == 1:
            return cls._tile_drop_directly(op)

        subset_df = in_df
        if op.subset:
            subset_df = in_df[op.subset]
        count_series = yield from recursive_tile(
            subset_df.agg("count", axis=1, _use_inf_as_na=op.use_inf_as_na)
        )

        nsplits, out_shape, left_chunks, right_chunks = align_dataframe_series(
            in_df, count_series, axis=0
        )
        out_chunk_indexes = itertools.product(*(range(s) for s in out_shape))

        out_chunks = []
        for out_idx, df_chunk in zip(out_chunk_indexes, left_chunks):
            series_chunk = right_chunks[out_idx[0]]
            kw = dict(
                shape=(np.nan, nsplits[1][out_idx[1]]),
                dtypes=df_chunk.dtypes,
                index_value=df_chunk.index_value,
                columns_value=df_chunk.columns_value,
            )

            new_op = op.copy().reset_key()
            new_op._drop_directly = False
            new_op._subset_size = len(op.subset) if op.subset else len(in_df.dtypes)
            out_chunks.append(
                new_op.new_chunk([df_chunk, series_chunk], index=out_idx, **kw)
            )

        new_op = op.copy().reset_key()
        params = out_df.params.copy()
        new_nsplits = list(tuple(ns) for ns in nsplits)
        new_nsplits[0] = (np.nan,) * len(new_nsplits[0])
        params.update(dict(nsplits=tuple(new_nsplits), chunks=out_chunks))
        return new_op.new_tileables(op.inputs, **params)

    @classmethod
    def execute(cls, ctx, op: "DataFrameDropNA"):
        try:
            pd.set_option("mode.use_inf_as_na", op.use_inf_as_na)

            in_data = ctx[op.inputs[0].key]
            if op.drop_directly:
                if isinstance(in_data, pd.DataFrame):
                    result = in_data.dropna(
                        axis=op.axis, how=op.how, thresh=op.thresh, subset=op.subset
                    )
                elif isinstance(in_data, pd.Series):
                    result = in_data.dropna(axis=op.axis, how=op.how)
                else:
                    result = in_data.dropna(how=op.how)
                ctx[op.outputs[0].key] = result
                return

            in_counts = ctx[op.inputs[1].key]
            if op.how == "all":
                in_counts = in_counts[in_counts > 0]
            else:
                thresh = op.subset_size if op.thresh is None else op.thresh
                in_counts = in_counts[in_counts >= thresh]

            ctx[op.outputs[0].key] = in_data.reindex(in_counts.index)
        finally:
            pd.reset_option("mode.use_inf_as_na")


def df_dropna(df, axis=0, how="any", thresh=None, subset=None, inplace=False):
    """
    Remove missing values.

    See the :ref:`User Guide <missing_data>` for more on which values are
    considered missing, and how to work with missing data.

    Parameters
    ----------
    axis : {0 or 'index', 1 or 'columns'}, default 0
        Determine if rows or columns which contain missing values are
        removed.

        * 0, or 'index' : Drop rows which contain missing values.
        * 1, or 'columns' : Drop columns which contain missing value.

        .. versionchanged:: 1.0.0

           Pass tuple or list to drop on multiple axes.
           Only a single axis is allowed.

    how : {'any', 'all'}, default 'any'
        Determine if row or column is removed from DataFrame, when we have
        at least one NA or all NA.

        * 'any' : If any NA values are present, drop that row or column.
        * 'all' : If all values are NA, drop that row or column.

    thresh : int, optional
        Require that many non-NA values.
    subset : array-like, optional
        Labels along other axis to consider, e.g. if you are dropping rows
        these would be a list of columns to include.
    inplace : bool, default False
        If True, do operation inplace and return None.

    Returns
    -------
    DataFrame
        DataFrame with NA entries dropped from it.

    See Also
    --------
    DataFrame.isna: Indicate missing values.
    DataFrame.notna : Indicate existing (non-missing) values.
    DataFrame.fillna : Replace missing values.
    Series.dropna : Drop missing values.
    Index.dropna : Drop missing indices.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({"name": ['Alfred', 'Batman', 'Catwoman'],
    ...                    "toy": [np.nan, 'Batmobile', 'Bullwhip'],
    ...                    "born": [md.NaT, md.Timestamp("1940-04-25"),
    ...                             md.NaT]})
    >>> df.execute()
           name        toy       born
    0    Alfred        NaN        NaT
    1    Batman  Batmobile 1940-04-25
    2  Catwoman   Bullwhip        NaT

    Drop the rows where at least one element is missing.

    >>> df.dropna().execute()
         name        toy       born
    1  Batman  Batmobile 1940-04-25

    Drop the rows where all elements are missing.

    >>> df.dropna(how='all').execute()
           name        toy       born
    0    Alfred        NaN        NaT
    1    Batman  Batmobile 1940-04-25
    2  Catwoman   Bullwhip        NaT

    Keep only the rows with at least 2 non-NA values.

    >>> df.dropna(thresh=2).execute()
           name        toy       born
    1    Batman  Batmobile 1940-04-25
    2  Catwoman   Bullwhip        NaT

    Define in which columns to look for missing values.

    >>> df.dropna(subset=['name', 'born']).execute()
           name        toy       born
    1    Batman  Batmobile 1940-04-25

    Keep the DataFrame with valid entries in the same variable.

    >>> df.dropna(inplace=True)
    >>> df.execute()
         name        toy       born
    1  Batman  Batmobile 1940-04-25
    """
    axis = validate_axis(axis, df)
    if axis != 0:
        raise NotImplementedError("Does not support dropna on DataFrame when axis=1")

    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameDropNA(
        axis=axis,
        how=how,
        thresh=thresh,
        subset=subset,
        output_types=[OutputType.dataframe],
        use_inf_as_na=use_inf_as_na,
    )
    out_df = op(df)
    if inplace:
        df.data = out_df.data
    else:
        return out_df


def series_dropna(series, axis=0, inplace=False, how=None):
    """
    Return a new Series with missing values removed.

    See the :ref:`User Guide <missing_data>` for more on which values are
    considered missing, and how to work with missing data.

    Parameters
    ----------
    axis : {0 or 'index'}, default 0
        There is only one axis to drop values from.
    inplace : bool, default False
        If True, do operation inplace and return None.
    how : str, optional
        Not in use. Kept for compatibility.

    Returns
    -------
    Series
        Series with NA entries dropped from it.

    See Also
    --------
    Series.isna: Indicate missing values.
    Series.notna : Indicate existing (non-missing) values.
    Series.fillna : Replace missing values.
    DataFrame.dropna : Drop rows or columns which contain NA values.
    Index.dropna : Drop missing indices.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> ser = md.Series([1., 2., np.nan])
    >>> ser.execute()
    0    1.0
    1    2.0
    2    NaN
    dtype: float64

    Drop NA values from a Series.

    >>> ser.dropna().execute()
    0    1.0
    1    2.0
    dtype: float64

    Keep the Series with valid entries in the same variable.

    >>> ser.dropna(inplace=True)
    >>> ser.execute()
    0    1.0
    1    2.0
    dtype: float64

    Empty strings are not considered NA values. ``None`` is considered an
    NA value.

    >>> ser = md.Series([np.NaN, 2, md.NaT, '', None, 'I stay'])
    >>> ser.execute()
    0       NaN
    1         2
    2       NaT
    3
    4      None
    5    I stay
    dtype: object
    >>> ser.dropna().execute()
    1         2
    3
    5    I stay
    dtype: object
    """
    axis = validate_axis(axis, series)
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameDropNA(
        axis=axis,
        how=how,
        output_types=[OutputType.series],
        use_inf_as_na=use_inf_as_na,
    )
    out_series = op(series)
    if inplace:
        series.data = out_series.data
    else:
        return out_series


def index_dropna(index, how="any"):
    """
    Return Index without NA/NaN values.

    Parameters
    ----------
    how : {'any', 'all'}, default 'any'
        If the Index is a MultiIndex, drop the value when any or all levels
        are NaN.

    Returns
    -------
    Index
    """
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameDropNA(
        axis=0, how=how, output_types=[OutputType.index], use_inf_as_na=use_inf_as_na
    )
    return op(index)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...serialization.serializables import (
    Serializable,
    FieldTypes,
    DictField,
    ReferenceField,
)
from ...utils import AttributeDict
from ..base import Base


class EntityData(Base):
    __slots__ = ("_siblings",)
    type_name = None

    # required fields
    _op = ReferenceField("op", "mars.core.operand.base.Operand")
    # optional fields
    _extra_params = DictField("extra_params", key_type=FieldTypes.string)

    def __init__(self, *args, **kwargs):
        extras = AttributeDict(
            (k, kwargs.pop(k)) for k in set(kwargs) - set(self._FIELDS)
        )
        kwargs["_extra_params"] = kwargs.pop("_extra_params", extras)
        super().__init__(*args, **kwargs)

    @property
    def op(self):
        return self._op

    @property
    def inputs(self):
        return self.op.inputs

    @inputs.setter
    def inputs(self, new_inputs):
        self.op.inputs = new_inputs

    def is_sparse(self):
        return self.op.is_sparse()

    issparse = is_sparse

    @property
    def extra_params(self):
        return self._extra_params

    def build_graph(self, **kw):
        from ..graph.builder.utils import build_graph

        return build_graph([self], **kw)

    def visualize(self, graph_attrs=None, node_attrs=None, **kw):
        from graphviz import Source

        g = self.build_graph(**kw)
        dot = g.to_dot(
            graph_attrs=graph_attrs,
            node_attrs=node_attrs,
            result_chunk_keys={c.key for c in self.chunks},
        )

        return Source(dot)

    def _need_execution(self):  # pylint: disable=no-self-use
        # some tileable may generate unknown meta,
        # they need to be executed first
        return False


class Entity(Serializable):
    _allow_data_type_ = ()
    type_name = None

    _data = ReferenceField("data", EntityData)

    def __init__(self, data=None, **kw):
        super().__init__(_data=data, **kw)

    def __dir__(self):
        obj_dir = object.__dir__(self)
        if self._data is not None:
            obj_dir = sorted(set(dir(self._data) + obj_dir))
        return obj_dir

    def __str__(self):
        return self._data.__str__()

    def __repr__(self):
        return self._data.__repr__()

    def _check_data(self, data):
        if data is not None and not isinstance(data, self._allow_data_type_):
            raise TypeError(f"Expect {self._allow_data_type_}, got {type(data)}")

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_data):
        self._check_data(new_data)
        self._data = new_data

    def __copy__(self):
        return self.copy()

    def copy(self):
        return self.copy_to(type(self)(None))

    def copy_to(self, target):
        target.data = self._data
        return target

    def copy_from(self, obj):
        self.data = obj.data

    def tiles(self):
        from .tileables import handler

        new_entity = self.copy()
        new_entity.data = handler.tiles(self.data)
        return new_entity

    def __getattr__(self, attr):
        return getattr(self._data, attr)

    def __setattr__(self, key, value):
        try:
            object.__setattr__(self, key, value)
        except AttributeError:
            return setattr(self._data, key, value)

    def _need_execution(self):
        return self._data._need_execution()


ENTITY_TYPE = (Entity, EntityData)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import numpy as np
from sklearn.exceptions import UndefinedMetricWarning

from ... import execute
from ... import tensor as mt
from ..utils.validation import (
    check_array,
    check_consistent_length,
    column_or_1d,
    _num_samples,
)


def _check_reg_targets(y_true, y_pred, multioutput, dtype="numeric"):
    """Check that y_true and y_pred belong to the same regression task.

    Parameters
    ----------
    y_true : array-like

    y_pred : array-like

    multioutput : array-like or string in ['raw_values', uniform_average',
        'variance_weighted'] or None
        None is accepted due to backward compatibility of r2_score().

    Returns
    -------
    type_true : one of {'continuous', continuous-multioutput'}
        The type of the true target data, as output by
        'utils.multiclass.type_of_target'.

    y_true : array-like of shape (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples, n_outputs)
        Estimated target values.

    multioutput : array-like of shape (n_outputs) or string in ['raw_values',
        uniform_average', 'variance_weighted'] or None
        Custom output weights if ``multioutput`` is array-like or
        just the corresponding argument if ``multioutput`` is a
        correct keyword.

    dtype : str or list, default="numeric"
        the dtype argument passed to check_array.
    """
    check_consistent_length(y_true, y_pred)
    y_true = check_array(y_true, ensure_2d=False, dtype=dtype)
    y_pred = check_array(y_pred, ensure_2d=False, dtype=dtype)

    if y_true.ndim == 1:
        y_true = y_true.reshape((-1, 1))

    if y_pred.ndim == 1:
        y_pred = y_pred.reshape((-1, 1))

    if y_true.shape[1] != y_pred.shape[1]:
        raise ValueError(
            "y_true and y_pred have different number of output "
            "({0}!={1})".format(y_true.shape[1], y_pred.shape[1])
        )

    n_outputs = y_true.shape[1]
    allowed_multioutput_str = ("raw_values", "uniform_average", "variance_weighted")
    if isinstance(multioutput, str):
        if multioutput not in allowed_multioutput_str:
            raise ValueError(
                "Allowed 'multioutput' string values are {}. "
                "You provided multioutput={!r}".format(
                    allowed_multioutput_str, multioutput
                )
            )
    elif multioutput is not None:
        multioutput = check_array(multioutput, ensure_2d=False)
        if n_outputs == 1:
            raise ValueError("Custom weights are useful only in multi-output cases.")
        elif n_outputs != len(multioutput):
            raise ValueError(
                ("There must be equally many custom weights (%d) as outputs (%d).")
                % (len(multioutput), n_outputs)
            )
    y_type = "continuous" if n_outputs == 1 else "continuous-multioutput"

    return y_type, y_true, y_pred, multioutput


def r2_score(
    y_true,
    y_pred,
    *,
    sample_weight=None,
    multioutput="uniform_average",
    session=None,
    run_kwargs=None
):
    """:math:`R^2` (coefficient of determination) regression score function.

    Best possible score is 1.0 and it can be negative (because the
    model can be arbitrarily worse). A constant model that always
    predicts the expected value of y, disregarding the input features,
    would get a :math:`R^2` score of 0.0.

    Read more in the :ref:`User Guide <r2_score>`.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    multioutput : {'raw_values', 'uniform_average', 'variance_weighted'}, \
            array-like of shape (n_outputs,) or None, default='uniform_average'

        Defines aggregating of multiple output scores.
        Array-like value defines weights used to average scores.
        Default is "uniform_average".

        'raw_values' :
            Returns a full set of scores in case of multioutput input.

        'uniform_average' :
            Scores of all outputs are averaged with uniform weight.

        'variance_weighted' :
            Scores of all outputs are averaged, weighted by the variances
            of each individual output.

    Returns
    -------
    z : float or tensor of floats
        The :math:`R^2` score or ndarray of scores if 'multioutput' is
        'raw_values'.

    Notes
    -----
    This is not a symmetric function.

    Unlike most other scores, :math:`R^2` score may be negative (it need not
    actually be the square of a quantity R).

    This metric is not well-defined for single samples and will return a NaN
    value if n_samples is less than two.

    References
    ----------
    .. [1] `Wikipedia entry on the Coefficient of determination
            <https://en.wikipedia.org/wiki/Coefficient_of_determination>`_

    Examples
    --------
    >>> from mars.learn.metrics import r2_score
    >>> y_true = [3, -0.5, 2, 7]
    >>> y_pred = [2.5, 0.0, 2, 8]
    >>> r2_score(y_true, y_pred)
    0.948...
    >>> y_true = [[0.5, 1], [-1, 1], [7, -6]]
    >>> y_pred = [[0, 2], [-1, 2], [8, -5]]
    >>> r2_score(y_true, y_pred,
    ...          multioutput='variance_weighted')
    0.938...
    >>> y_true = [1, 2, 3]
    >>> y_pred = [1, 2, 3]
    >>> r2_score(y_true, y_pred)
    1.0
    >>> y_true = [1, 2, 3]
    >>> y_pred = [2, 2, 2]
    >>> r2_score(y_true, y_pred)
    0.0
    >>> y_true = [1, 2, 3]
    >>> y_pred = [3, 2, 1]
    >>> r2_score(y_true, y_pred)
    -3.0
    """
    _, y_true, y_pred, multioutput = _check_reg_targets(y_true, y_pred, multioutput)
    check_consistent_length(y_true, y_pred, sample_weight)

    if _num_samples(y_pred) < 2:
        msg = "R^2 score is not well-defined with less than two samples."
        warnings.warn(msg, UndefinedMetricWarning)
        return float("nan")

    if sample_weight is not None:
        sample_weight = column_or_1d(sample_weight)
        weight = sample_weight[:, np.newaxis]
    else:
        weight = 1.0

    numerator = (weight * (y_true - y_pred) ** 2).sum(axis=0, dtype=np.float64)
    denominator = (
        weight * (y_true - mt.average(y_true, axis=0, weights=sample_weight)) ** 2
    ).sum(axis=0, dtype=np.float64)
    nonzero_denominator = denominator != 0
    nonzero_numerator = numerator != 0
    valid_score = nonzero_denominator & nonzero_numerator
    output_scores = mt.ones((y_true.shape[1],))
    output_scores[valid_score] = 1 - (numerator[valid_score] / denominator[valid_score])
    # arbitrary set to zero to avoid -inf scores, having a constant
    # y_true is not interesting for scoring a regression anyway
    output_scores[nonzero_numerator & ~nonzero_denominator] = 0.0
    if isinstance(multioutput, str):
        if multioutput == "raw_values":
            # return scores individually
            return output_scores
        elif multioutput == "uniform_average":
            # passing None as weights results is uniform mean
            avg_weights = None
        elif multioutput == "variance_weighted":
            avg_weights = denominator
            # avoid fail on constant y or one-element arrays
            cond1 = mt.any(nonzero_denominator)
            execute(
                cond1, nonzero_denominator, session=session, **(run_kwargs or dict())
            )
            if not cond1.fetch():
                if not mt.any(nonzero_numerator).to_numpy(
                    session=session, **(run_kwargs or dict())
                ):
                    return 1.0
                else:
                    return 0.0
    else:
        avg_weights = multioutput

    return mt.average(output_scores, weights=avg_weights).execute(
        session=session, **(run_kwargs or dict())
    )



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def setdiff1d(ar1, ar2, assume_unique=False):
    """
    Find the set difference of two tensors.

    Return the unique values in `ar1` that are not in `ar2`.

    Parameters
    ----------
    ar1 : array_like
        Input tensor.
    ar2 : array_like
        Input comparison tensor.
    assume_unique : bool
        If True, the input tensors are both assumed to be unique, which
        can speed up the calculation.  Default is False.

    Returns
    -------
    setdiff1d : Tensor
        1D tensor of values in `ar1` that are not in `ar2`. The result
        is sorted when `assume_unique=False`, but otherwise only sorted
        if the input is sorted.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.array([1, 2, 3, 2, 4, 1])
    >>> b = mt.array([3, 4, 5, 6])
    >>> mt.setdiff1d(a, b).execute()
    array([1, 2])

    """

    from ..datasource.array import asarray
    from .in1d import in1d
    from .unique import unique

    if assume_unique:
        ar1 = asarray(ar1).ravel()
    else:
        ar1 = unique(ar1)
        ar2 = unique(ar2)
    return ar1[in1d(ar1, ar2, assume_unique=True, invert=True)]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import log

try:
    from scipy.stats import entropy as sp_entropy
except ImportError:
    sp_entropy = None

from ... import tensor as mt
from ...tensor import special as mt_special
from ..core import TENSOR_TYPE
from ..datasource import tensor as astensor
from ..utils import implement_scipy


@implement_scipy(sp_entropy)
def entropy(pk, qk=None, base=None):
    pk = astensor(pk)
    pk = 1.0 * pk / mt.sum(pk, axis=0)
    if qk is None:
        vec = mt_special.entr(pk)
    else:
        qk = astensor(qk)
        if len(qk) != len(pk):
            raise ValueError("qk and pk must have same length.")
        qk = 1.0 * qk / mt.sum(qk, axis=0)
        vec = mt_special.rel_entr(pk, qk)
    S = mt.sum(vec, axis=0)
    if base is not None:
        if isinstance(base, TENSOR_TYPE):
            S /= mt.log(base)
        else:
            S /= log(base)
    return S



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .core import Serializable, SerializableMeta
from .field import (
    AnyField,
    IdentityField,
    BoolField,
    Int8Field,
    Int16Field,
    Int32Field,
    Int64Field,
    UInt8Field,
    UInt16Field,
    UInt32Field,
    UInt64Field,
    Float16Field,
    Float32Field,
    Float64Field,
    Complex64Field,
    Complex128Field,
    StringField,
    BytesField,
    KeyField,
    NDArrayField,
    Datetime64Field,
    Timedelta64Field,
    DataTypeField,
    IndexField,
    SeriesField,
    DataFrameField,
    IntervalArrayField,
    SliceField,
    FunctionField,
    NamedTupleField,
    TZInfoField,
    ListField,
    TupleField,
    DictField,
    ReferenceField,
    OneOfField,
)
from .field_type import FieldTypes



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes
from ...serialization.serializables import KeyField, Float64Field, Int64Field
from ...tensor.base.rebalance import RebalanceMixin
from ..core import INDEX_TYPE
from ..initializer import DataFrame as asdataframe, Series as asseries, Index as asindex
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import validate_axis


class DataFrameRebalance(RebalanceMixin, DataFrameOperandMixin, DataFrameOperand):
    _op_type_ = opcodes.REBALANCE

    _input = KeyField("input")
    _factor = Float64Field("factor")
    _axis = Int64Field("axis")
    _num_partitions = Int64Field("num_partitions")

    def __init__(
        self,
        input=None,
        factor=None,
        axis=None,  # pylint: disable=redefined-builtin
        num_partitions=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _input=input,
            _factor=factor,
            _axis=axis,
            _num_partitions=num_partitions,
            _output_types=output_types,
            **kw
        )

    @property
    def input(self):
        return self._input

    @property
    def factor(self):
        return self._factor

    @property
    def axis(self):
        return self._axis

    @property
    def num_partitions(self):
        return self._num_partitions

    def _get_input_object(self):
        in_obj = self.input
        if isinstance(in_obj, INDEX_TYPE):
            convert = asindex
        else:
            convert = asdataframe if in_obj.ndim == 2 else asseries
        return convert(in_obj)


def rebalance(
    df_or_series, factor=None, axis=0, num_partitions=None, reassign_worker=True
):
    """
    Make Data more balanced across entire cluster.

    Parameters
    ----------
    factor : float
        Specified so that number of chunks after balance is
        total CPU count of cluster * factor.
    axis : int
        The axis to rebalance.
    num_partitions : int
        Specified so the number of chunks are at most
        num_partitions.
    reassign_worker : bool
        If True, workers will be reassigned.

    Returns
    -------
    Series or DataFrame
        Result of DataFrame or Series after rebalanced.
    """
    axis = validate_axis(axis, df_or_series)
    if num_partitions is None:
        factor = factor if factor is not None else 1.2

    op = DataFrameRebalance(
        input=df_or_series,
        factor=factor,
        axis=axis,
        num_partitions=num_partitions,
        reassign_worker=reassign_worker,
    )
    return op(df_or_series)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes
from ...serialization.serializables import ReferenceField
from ..entity import FuseChunk, FuseChunkData, NotSupportTile
from ..graph import ChunkGraph
from .base import Operand


class Fuse(Operand):
    __slots__ = ("_fuse_graph",)
    _op_type_ = opcodes.FUSE

    fuse_graph = ReferenceField("fuse_graph", ChunkGraph)


class FuseChunkMixin:
    __slots__ = ()

    def _create_chunk(self, output_idx, index, **kw):
        data = FuseChunkData(_index=index, _op=self, **kw)
        return FuseChunk(data)

    @classmethod
    def tile(cls, op):
        raise NotSupportTile("FuseChunk is a chunk operand which does not support tile")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="binary_or")
class TensorAnd(TensorBinOp):
    _op_type_ = OperandDef.AND
    _func_name = "logical_and"


@infer_dtype(np.logical_and)
def logical_and(x1, x2, out=None, where=None, **kwargs):
    """
    Compute the truth value of x1 AND x2 element-wise.

    Parameters
    ----------
    x1, x2 : array_like
        Input tensors. `x1` and `x2` must be of the same shape.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor or bool
        Boolean result with the same shape as `x1` and `x2` of the logical
        AND operation on corresponding elements of `x1` and `x2`.

    See Also
    --------
    logical_or, logical_not, logical_xor
    bitwise_and

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.logical_and(True, False).execute()
    False
    >>> mt.logical_and([True, False], [False, False]).execute()
    array([False, False])

    >>> x = mt.arange(5)
    >>> mt.logical_and(x>1, x<4).execute()
    array([False, False,  True,  True, False])
    """
    op = TensorAnd(**kwargs)
    return op(x1, x2, out=out, where=where)


@infer_dtype(np.logical_and, reverse=True)
def rlogical_and(x1, x2, **kwargs):
    op = TensorAnd(**kwargs)
    return op.rcall(x1, x2)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ....lib.aio import alru_cache
from ..core import NodeRole
from ..locator import SupervisorLocatorActor


class SupervisorPeerLocatorActor(SupervisorLocatorActor):
    _node_role = NodeRole.SUPERVISOR

    @classmethod
    def default_uid(cls):
        return SupervisorLocatorActor.__name__

    async def __post_create__(self):
        await super().__post_create__()

        supervisors = await self._backend.get_supervisors(filter_ready=False)
        try:
            node_info_ref = await self._get_node_info_ref()
            await node_info_ref.put_starting_nodes(supervisors, NodeRole.SUPERVISOR)
        except mo.ActorNotExist:  # pragma: no cover
            pass

    @alru_cache(cache_exceptions=False)
    async def _get_node_info_ref(self):
        from .node_info import NodeInfoCollectorActor

        return await mo.actor_ref(
            uid=NodeInfoCollectorActor.default_uid(), address=self.address
        )

    async def _get_supervisors_from_backend(self, filter_ready: bool = True):
        return await self._backend.get_supervisors(filter_ready=filter_ready)

    async def _watch_supervisors_from_backend(self):
        async for supervisors in self._backend.watch_supervisors():
            yield supervisors



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import warnings

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import (
    AnyField,
    BoolField,
    ListField,
    Int64Field,
    ReferenceField,
)
from ..utils import parse_index, lazy_import, tokenize
from .core import (
    IncrementalIndexDatasource,
    IncrementalIndexDataSourceMixin,
    HeadOptimizedDataSource,
)

ray = lazy_import("ray")
# Ray Datasets is available in early preview at ray.data with Ray 1.6+
# (and ray.experimental.data in Ray 1.5)
ray_dataset = lazy_import("ray.data", rename="ray_dataset")
ray_exp_dataset = lazy_import("ray.experimental.data", rename="ray_exp_dataset")
real_ray_dataset = ray_dataset or ray_exp_dataset


class DataFrameReadRayDataset(
    IncrementalIndexDatasource, IncrementalIndexDataSourceMixin
):
    _op_type_ = OperandDef.READ_RAYDATASET

    refs = AnyField("refs", default=None)
    columns = ListField("columns", default=None)
    incremental_index = BoolField("incremental_index", default=None)
    nrows = Int64Field("nrows", default=None)

    @classmethod
    def _tile_partitioned(cls, op: "DataFrameReadRayDataset"):
        out_df = op.outputs[0]
        shape = (np.nan, out_df.shape[1])
        dtypes = out_df.dtypes
        dataset = op.refs

        chunk_index = 0
        out_chunks = []
        for object_ref in dataset:
            chunk_op = op.copy().reset_key()
            chunk_op._refs = [object_ref]
            new_chunk = chunk_op.new_chunk(
                None,
                shape=shape,
                index=(chunk_index, 0),
                index_value=out_df.index_value,
                columns_value=out_df.columns_value,
                dtypes=dtypes,
            )
            out_chunks.append(new_chunk)
            chunk_index += 1

        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_df.shape[1],))
        return new_op.new_dataframes(
            None,
            out_df.shape,
            dtypes=dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile(cls, op):
        return cls._tile_partitioned(op)

    @classmethod
    def execute(cls, ctx, op: "DataFrameReadRayDataset"):
        out = op.outputs[0]
        ref = op.refs[0]

        df = ray.get(ref)
        ctx[out.key] = df

    def __call__(self, index_value=None, columns_value=None, dtypes=None):
        shape = (np.nan, len(dtypes))
        return self.new_dataframe(
            None,
            shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )


def read_ray_dataset(ds, columns=None, incremental_index=False, **kwargs):
    assert isinstance(ds, real_ray_dataset.Dataset)
    refs = ds.to_pandas_refs()
    schema = ds.schema()

    import pyarrow as pa

    try:
        from ray.data.impl.pandas_block import PandasBlockSchema

        if isinstance(schema, PandasBlockSchema):
            dtypes = pd.Series(schema.types, index=schema.names)
        elif isinstance(schema, pa.Schema):
            dtypes = schema.empty_table().to_pandas().dtypes
        else:
            raise NotImplementedError(f"Unsupported format of schema {schema}")
    except ImportError:  # pragma: no cover
        dtypes = schema.empty_table().to_pandas().dtypes
    index_value = parse_index(pd.RangeIndex(-1))
    columns_value = parse_index(dtypes.index, store_data=True)
    op = DataFrameReadRayDataset(
        refs=refs, columns=columns, incremental_index=incremental_index
    )
    return op(index_value=index_value, columns_value=columns_value, dtypes=dtypes)


# keep it for back compatibility
@functools.wraps(read_ray_dataset)
def read_raydataset(*args, **kwargs):
    warnings.warn(
        "read_raydataset has been renamed to read_ray_dataset",
        DeprecationWarning,
    )
    return read_ray_dataset(*args, **kwargs)


class DataFrameReadMLDataset(HeadOptimizedDataSource):
    _op_type_ = OperandDef.READ_MLDATASET

    mldataset = ReferenceField("mldataset", "ray.util.data.MLDataset", default=None)
    columns = ListField("columns", default=None)

    def __init__(self, **kw):
        super().__init__(_output_types=[OutputType.dataframe], **kw)

    def _update_key(self):
        """We can't direct generate token for mldataset when we use
        ray client, so we use all mldataset's actor_id to generate
        token.
        """
        datas = []
        for value in self._values_:
            if isinstance(value, ray.util.data.MLDataset):
                actor_sets = [
                    ([str(actor) for actor in actor_set.actors], actor_set.transforms)
                    for actor_set in value.actor_sets
                ]
                datas.append(actor_sets)
                continue
            datas.append(value)
        self._obj_set("_key", tokenize(type(self).__name__, *datas))
        return self

    def __call__(self, dtypes, nrows: int):
        columns_value = parse_index(dtypes.index, store_data=True)
        index_value = parse_index(pd.RangeIndex(nrows))
        return self.new_dataframe(
            None,
            (nrows, len(dtypes)),
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    @classmethod
    def tile(cls, op: "DataFrameReadMLDataset"):
        count_iter = op.mldataset.for_each(lambda df: len(df))
        nsplits = [sum(shard) for shard in count_iter.shards()]
        nsplits_acc = np.cumsum(nsplits)
        out_df = op.outputs[0]
        out_chunks = []
        for shard_index in range(op.mldataset.num_shards()):
            chunk_op = op.copy().reset_key()
            # Make chunk key unique, otherwise all chunk will have same key.
            # See `DataFrameFromRecords#tile`
            chunk_op.extra_params["shard_index"] = shard_index
            shape = (nsplits[shard_index], out_df.shape[1])
            begin_index = nsplits_acc[shard_index] - nsplits[shard_index]
            end_index = nsplits_acc[shard_index]
            index = parse_index(pd.RangeIndex(start=begin_index, stop=end_index))
            new_chunk = chunk_op.new_chunk(
                None,
                shape=shape,
                index=(shard_index, 0),
                index_value=index,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
            )
            out_chunks.append(new_chunk)
        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_df.shape[1],))
        return new_op.new_dataframes(
            None,
            out_df.shape,
            dtypes=out_df.dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        shard = op.mldataset.get_shard(chunk.index[0])
        pd_dfs = list(shard)
        pd_df = pd.concat(pd_dfs).set_index(chunk.index_value.to_pandas())
        ctx[chunk.key] = pd_df


def read_ray_mldataset(mldataset, **kwargs):
    import ray.util.data

    assert isinstance(mldataset, ray.util.data.MLDataset)
    not_empty_dfs = mldataset.filter(lambda df: len(df) > 0).take(1)
    if not not_empty_dfs:
        raise ValueError(
            f"MLDataset {mldataset} is empty, please provide an non-empty dataset."
        )
    df_record: pd.DataFrame = not_empty_dfs[0]
    columns = df_record.columns.names
    nrows = sum(mldataset.for_each(lambda df: len(df)).gather_async())
    op = DataFrameReadMLDataset(mldataset=mldataset, columns=columns, nrows=nrows)
    return op(df_record.dtypes, nrows)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import base64
import json
from typing import Callable, List, Optional, Union

from ....core import TileableGraph, Tileable
from ....lib.tbcode import load_traceback_code, dump_traceback_code
from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import TaskResult, TaskStatus
from .core import AbstractTaskAPI


def _json_serial_task_result(result: Optional[TaskResult]):
    if result is None:
        return {}
    res_json = {
        "task_id": result.task_id,
        "session_id": result.session_id,
        "stage_id": result.stage_id,
        "start_time": result.start_time,
        "end_time": result.end_time,
        "progress": result.progress,
        "status": result.status.value,
        "profiling": result.profiling,
    }
    if result.error is not None:
        res_json["error"] = base64.b64encode(
            serialize_serializable(result.error)
        ).decode()
        res_json["traceback"] = base64.b64encode(
            serialize_serializable(result.traceback)
        ).decode()
        res_json["traceback_code"] = dump_traceback_code(result.traceback)
    return res_json


def _json_deserial_task_result(d: dict) -> Optional[TaskResult]:
    if not d:
        return None
    if "error" in d:
        d["error"] = deserialize_serializable(base64.b64decode(d["error"]))
        d["traceback"] = deserialize_serializable(base64.b64decode(d["traceback"]))
        load_traceback_code(d.pop("traceback_code"))
    d["status"] = TaskStatus(d["status"])
    return TaskResult(**d)


class TaskWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = "/api/session/(?P<session_id>[^/]+)/task"

    async def _get_oscar_task_api(self, session_id: str):
        from .oscar import TaskAPI

        return await self._get_api_by_key(TaskAPI, session_id)

    @web_api("", method="post")
    async def submit_tileable_graph(self, session_id: str):
        body_args = (
            deserialize_serializable(self.request.body) if self.request.body else None
        )

        task_name = body_args.get("task_name", None) or None
        fuse_enabled = body_args.get("fuse")

        graph = body_args["graph"]
        extra_config = body_args.get("extra_config", None)
        if extra_config:
            extra_config = deserialize_serializable(extra_config)

        oscar_api = await self._get_oscar_task_api(session_id)
        task_id = await oscar_api.submit_tileable_graph(
            graph,
            task_name=task_name,
            fuse_enabled=fuse_enabled,
            extra_config=extra_config,
        )
        self.write(task_id)

    @web_api("", method="get", cache_blocking=True)
    async def get_task_results(self, session_id: str):
        progress = bool(int(self.get_argument("progress", "0")))
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_task_results(progress=progress)
        self.write(json.dumps({"tasks": [_json_serial_task_result(r) for r in res]}))

    @web_api(
        "(?P<task_id>[^/]+)",
        method="get",
        arg_filter={"action": "fetch_tileables"},
        cache_blocking=True,
    )
    async def get_fetch_tileables(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_fetch_tileables(task_id)
        self.write(serialize_serializable(res))

    @web_api("(?P<task_id>[^/]+)", method="get", cache_blocking=True)
    async def get_task_result(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_task_result(task_id)
        self.write(json.dumps(_json_serial_task_result(res)))

    @web_api(
        "(?P<task_id>[^/]+)/tileable_graph",
        method="get",
        arg_filter={"action": "get_tileable_graph_as_json"},
        cache_blocking=True,
    )
    async def get_tileable_graph_as_json(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_tileable_graph_as_json(task_id)
        self.write(json.dumps(res))

    @web_api("(?P<task_id>[^/]+)/tileable_detail", method="get", cache_blocking=True)
    async def get_tileable_details(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_tileable_details(task_id)
        self.write(json.dumps(res))

    @web_api(
        "(?P<task_id>[^/]+)/(?P<tileable_id>[^/]+)/subtask",
        method="get",
        cache_blocking=True,
    )
    async def get_tileable_subtasks(
        self, session_id: str, task_id: str, tileable_id: str
    ):
        with_input_output = self.get_argument("with_input_output", "false") == "true"
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_tileable_subtasks(
            task_id, tileable_id, with_input_output
        )
        self.write(json.dumps(res))

    @web_api(
        "(?P<task_id>[^/]+)",
        method="get",
        arg_filter={"action": "progress"},
        cache_blocking=True,
    )
    async def get_task_progress(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_task_progress(task_id)
        self.write(str(res))

    @web_api("", method="get", arg_filter={"action": "last_idle_time"})
    async def get_last_idle_time(self, session_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_last_idle_time()
        if res:
            self.write(str(res))

    @web_api("(?P<task_id>[^/]+)", method="get", arg_filter={"action": "wait"})
    async def wait_task(self, session_id: str, task_id: str):
        timeout = self.get_argument("timeout", None) or None
        timeout = float(timeout) if timeout is not None else None
        oscar_api = await self._get_oscar_task_api(session_id)
        if timeout:
            try:
                res = await asyncio.wait_for(
                    asyncio.shield(oscar_api.wait_task(task_id, timeout)),
                    timeout=timeout,
                )
                self.write(json.dumps(_json_serial_task_result(res)))
            except asyncio.TimeoutError:
                self.write(json.dumps({}))
        else:
            res = await oscar_api.wait_task(task_id, timeout)
            self.write(json.dumps(_json_serial_task_result(res)))

    @web_api("(?P<task_id>[^/]+)", method="delete")
    async def cancel_task(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        await oscar_api.cancel_task(task_id)


web_handlers = {TaskWebAPIHandler.get_root_pattern(): TaskWebAPIHandler}


class WebTaskAPI(AbstractTaskAPI, MarsWebAPIClientMixin):
    def __init__(
        self, session_id: str, address: str, request_rewriter: Callable = None
    ):
        self._session_id = session_id
        self._address = address.rstrip("/")
        self.request_rewriter = request_rewriter

    async def get_task_results(self, progress: bool = False) -> List[TaskResult]:
        path = f"{self._address}/api/session/{self._session_id}/task"
        params = {"progress": int(progress)}
        res = await self._request_url("GET", path, params=params)
        return [
            _json_deserial_task_result(d)
            for d in json.loads(res.body.decode())["tasks"]
        ]

    async def submit_tileable_graph(
        self,
        graph: TileableGraph,
        task_name: str = None,
        fuse_enabled: bool = True,
        extra_config: dict = None,
    ) -> str:
        path = f"{self._address}/api/session/{self._session_id}/task"
        extra_config_ser = (
            serialize_serializable(extra_config) if extra_config else None
        )
        body = serialize_serializable(
            {
                "task_name": task_name if task_name else "",
                "fuse": fuse_enabled,
                "graph": graph,
                "extra_config": extra_config_ser,
            }
        )
        res = await self._request_url(
            path=path,
            method="POST",
            headers={"Content-Type": "application/octet-stream"},
            data=body,
        )
        return res.body.decode().strip()

    async def get_fetch_tileables(self, task_id: str) -> List[Tileable]:
        path = (
            f"{self._address}/api/session/{self._session_id}/task/{task_id}"
            f"?action=fetch_tileables"
        )
        res = await self._request_url("GET", path)
        return deserialize_serializable(res.body)

    async def get_task_result(self, task_id: str) -> TaskResult:
        path = f"{self._address}/api/session/{self._session_id}/task/{task_id}"
        res = await self._request_url("GET", path)
        return _json_deserial_task_result(json.loads(res.body.decode()))

    async def get_task_progress(self, task_id: str) -> float:
        path = f"{self._address}/api/session/{self._session_id}/task/{task_id}"
        params = dict(action="progress")
        res = await self._request_url("GET", path, params=params)
        return float(res.body.decode())

    async def get_last_idle_time(self) -> Union[float, None]:
        path = f"{self._address}/api/session/{self._session_id}/task"
        params = dict(action="last_idle_time")
        res = await self._request_url("GET", path, params=params)
        content = res.body.decode()
        return float(content) if content else None

    async def wait_task(self, task_id: str, timeout: float = None):
        path = f"{self._address}/api/session/{self._session_id}/task/{task_id}"
        # increase client timeout to handle network overhead during entire request
        client_timeout = timeout + 3 if timeout else 0
        params = {"action": "wait", "timeout": "" if timeout is None else str(timeout)}
        res = await self._request_url(
            "GET", path, params=params, request_timeout=client_timeout
        )
        return _json_deserial_task_result(json.loads(res.body.decode()))

    async def cancel_task(self, task_id: str):
        path = f"{self._address}/api/session/{self._session_id}/task/{task_id}"
        await self._request_url(path=path, method="DELETE")

    async def get_tileable_graph_as_json(self, task_id: str):
        path = f"{self._address}/api/session/{self._session_id}/task/{task_id}/tileable_graph"
        params = dict(action="get_tileable_graph_as_json")
        res = await self._request_url(path=path, params=params, method="GET")
        return json.loads(res.body.decode())

    async def get_tileable_details(self, task_id: str):
        path = f"{self._address}/api/session/{self._session_id}/task/{task_id}/tileable_detail"
        res = await self._request_url(path=path, method="GET")
        return json.loads(res.body.decode())

    async def get_tileable_subtasks(
        self, task_id: str, tileable_id: str, with_input_output: bool
    ):

        with_input_output = "true" if with_input_output else "false"
        path = f"{self._address}/api/session/{self._session_id}/task/{task_id}/{tileable_id}/subtask"
        params = {
            "action": "fetch_graph",
            "with_input_output": with_input_output,
        }
        res = await self._request_url(path=path, params=params, method="GET")
        return json.loads(res.body.decode())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import itertools
import numpy as np
from collections import defaultdict
from typing import Dict, List, Set

from .... import oscar as mo
from ....core.operand import Fetch, FetchShuffle
from ....typing import BandType
from ...core import NodeRole
from ...subtask import Subtask
from ..errors import NoMatchingSlots, NoAvailableBand


class AssignerActor(mo.Actor):
    _bands: List[BandType]

    @classmethod
    def gen_uid(cls, session_id: str):
        return f"{session_id}_assigner"

    def __init__(self, session_id: str):
        self._session_id = session_id
        self._slots_ref = None

        self._cluster_api = None
        self._meta_api = None

        self._bands = []
        self._address_to_bands = dict()
        self._device_type_to_bands = dict()
        self._band_watch_task = None

    async def __post_create__(self):
        from ...cluster.api import ClusterAPI
        from ...meta.api import MetaAPI

        self._cluster_api = await ClusterAPI.create(self.address)
        self._meta_api = await MetaAPI.create(
            session_id=self._session_id, address=self.address
        )

        async def watch_bands():
            async for bands in self._cluster_api.watch_all_bands(NodeRole.WORKER):
                self._update_bands(list(bands))

        self._band_watch_task = asyncio.create_task(watch_bands())

    async def __pre_destroy__(self):
        if self._band_watch_task is not None:  # pragma: no branch
            self._band_watch_task.cancel()

    def _update_bands(self, bands: List[BandType]):
        self._bands = bands

        grouped_bands = itertools.groupby(sorted(self._bands), key=lambda b: b[0])
        self._address_to_bands = {k: list(v) for k, v in grouped_bands}

        grouped_bands = itertools.groupby(
            sorted(("numa" if b[1].startswith("numa") else "gpu", b) for b in bands),
            key=lambda tp: tp[0],
        )
        self._device_type_to_bands = {
            k: [v[1] for v in tps] for k, tps in grouped_bands
        }

    def _get_device_bands(self, is_gpu: bool):
        band_prefix = "numa" if not is_gpu else "gpu"
        filtered_bands = self._device_type_to_bands.get(band_prefix) or []
        if not filtered_bands:
            raise NoMatchingSlots("gpu" if is_gpu else "cpu")
        return filtered_bands

    def _get_random_band(
        self,
        is_gpu: bool,
        exclude_bands: Set[BandType] = None,
        random_when_unavailable: bool = True,
    ):
        bands = self._get_device_bands(is_gpu)
        if exclude_bands:
            avail_bands = [band for band in bands if band not in exclude_bands]
            if avail_bands:
                return avail_bands[np.random.choice(len(avail_bands))]
            elif not random_when_unavailable:
                raise NoAvailableBand(
                    f"No bands available after excluding bands {exclude_bands}"
                )
        return bands[np.random.choice(len(bands))]

    async def assign_subtasks(
        self,
        subtasks: List[Subtask],
        exclude_bands: Set[BandType] = None,
        random_when_unavailable: bool = True,
    ):
        exclude_bands = exclude_bands or set()
        inp_keys = set()
        broadcaster_keys = set()
        selected_bands = dict()

        if not self._bands:
            self._update_bands(
                list(await self._cluster_api.get_all_bands(NodeRole.WORKER))
            )

        for subtask in subtasks:
            is_gpu = any(c.op.gpu for c in subtask.chunk_graph)
            if subtask.expect_bands:
                # exclude expected but unready bands
                expect_available_bands = [
                    expect_band
                    for expect_band in subtask.expect_bands
                    if expect_band in self._bands and expect_band not in exclude_bands
                ]
                # fill in if all expected bands are unready
                if not expect_available_bands:
                    expect_available_bands = [
                        self._get_random_band(
                            is_gpu, exclude_bands, random_when_unavailable
                        )
                    ]
                selected_bands[subtask.subtask_id] = expect_available_bands
                continue
            for indep_chunk in subtask.chunk_graph.iter_indep():
                if isinstance(indep_chunk.op, Fetch):
                    if indep_chunk.is_broadcaster:
                        broadcaster_keys.add(indep_chunk.key)
                    inp_keys.add(indep_chunk.key)
                elif isinstance(indep_chunk.op, FetchShuffle):
                    selected_bands[subtask.subtask_id] = [
                        self._get_random_band(
                            is_gpu, exclude_bands, random_when_unavailable
                        )
                    ]
                    break

        fields = ["store_size", "bands"]
        inp_keys = list(inp_keys)
        metas = await self._meta_api.get_chunk_meta.batch(
            *(self._meta_api.get_chunk_meta.delay(key, fields) for key in inp_keys)
        )

        inp_metas = dict(zip(inp_keys, metas))
        if broadcaster_keys:
            # set broadcaster's size as 0 to avoid assigning all successors to same band.
            for key in broadcaster_keys:
                inp_metas[key]["store_size"] = 0
        assigns = []
        for subtask in subtasks:
            is_gpu = any(c.op.gpu for c in subtask.chunk_graph)
            band_prefix = "numa" if not is_gpu else "gpu"
            filtered_bands = self._get_device_bands(is_gpu)

            if subtask.subtask_id in selected_bands:
                bands = selected_bands[subtask.subtask_id]
            else:
                band_sizes = defaultdict(lambda: 0)
                for inp in subtask.chunk_graph.iter_indep():
                    if not isinstance(inp.op, Fetch):  # pragma: no cover
                        continue
                    meta = inp_metas[inp.key]
                    for band in meta["bands"]:
                        if not band[1].startswith(band_prefix):
                            sel_bands = [
                                b
                                for b in self._address_to_bands[band[0]]
                                if b[1].startswith(band_prefix)
                                and b not in exclude_bands
                            ]
                            if sel_bands:
                                band = sel_bands[np.random.choice(len(sel_bands))]
                        if band not in filtered_bands or band in exclude_bands:
                            band = self._get_random_band(
                                is_gpu, exclude_bands, random_when_unavailable
                            )
                        band_sizes[band] += meta["store_size"]
                bands = []
                max_size = -1
                for band, size in band_sizes.items():
                    if size > max_size:
                        bands = [band]
                        max_size = size
                    elif size == max_size:
                        bands.append(band)
            band = bands[np.random.choice(len(bands))]
            if (
                not random_when_unavailable and band in exclude_bands
            ):  # pragma: no cover
                raise NoAvailableBand(
                    f"No bands available for subtask {subtask.subtask_id} after "
                    f"excluded {exclude_bands}"
                )
            if subtask.bands_specified and band not in subtask.expect_bands:
                raise NoAvailableBand(
                    f"No bands available for subtask {subtask.subtask_id} on bands {subtask.expect_bands} "
                    f"after excluded {exclude_bands}"
                )
            assigns.append(band)
        return assigns

    async def reassign_subtasks(
        self, band_to_queued_num: Dict[BandType, int]
    ) -> Dict[BandType, int]:
        move_queued_subtasks = {}
        for is_gpu in (False, True):
            band_name_prefix = "numa" if not is_gpu else "gpu"

            filtered_bands = [
                band for band in self._bands if band[1].startswith(band_name_prefix)
            ]
            filtered_band_to_queued_num = {
                k: v
                for k, v in band_to_queued_num.items()
                if k[1].startswith(band_name_prefix)
            }

            if not filtered_bands:
                continue

            num_used_bands = len(filtered_band_to_queued_num.keys())
            if num_used_bands == 1:
                [(band, length)] = filtered_band_to_queued_num.items()
                if length == 0:
                    move_queued_subtasks.update({band: 0})
                    continue
                # no need to balance when there's only one band initially
                if len(filtered_bands) == 1 and band == filtered_bands[0]:
                    move_queued_subtasks.update({band: 0})
                    continue
            # unready bands recorded in band_num_queued_subtasks, some of them may hold 0 subtasks
            unready_bands = list(
                set(filtered_band_to_queued_num.keys()) - set(filtered_bands)
            )
            # ready bands not recorded in band_num_queued_subtasks, all of them hold 0 subtasks
            new_ready_bands = list(
                set(filtered_bands) - set(filtered_band_to_queued_num.keys())
            )
            # when there are new ready bands, make all bands hold same amount of subtasks
            # when there are no new ready bands now, move out subtasks left on them
            if not new_ready_bands and unready_bands:
                filtered_band_to_queued_num = {
                    k: filtered_band_to_queued_num[k] for k in unready_bands
                }
            # approximate total of subtasks moving to each ready band
            num_all_subtasks = sum(filtered_band_to_queued_num.values())
            mean = int(num_all_subtasks / len(filtered_bands))
            # all_bands (namely) includes:
            # a. ready bands recorded in band_num_queued_subtasks
            # b. ready bands not recorded in band_num_queued_subtasks
            # c. unready bands recorded in band_num_queued_subtasks
            # a. + b. = self._bands, a. + c. = bands in band_num_queued_subtasks
            all_bands = list(
                set(filtered_bands) | set(filtered_band_to_queued_num.keys())
            )
            # calculate the differential steps of moving subtasks
            # move < 0 means subtasks should move out and vice versa
            # unready bands no longer hold subtasks
            # assuming bands not recorded in band_num_queued_subtasks hold 0 subtasks
            band_move_nums = {}
            for band in all_bands:
                if band in filtered_bands:
                    band_move_nums[band] = mean - filtered_band_to_queued_num.get(
                        band, 0
                    )
                else:
                    band_move_nums[band] = -filtered_band_to_queued_num.get(band, 0)
            # ensure the balance of moving in and out
            total_move = sum(band_move_nums.values())
            # int() is going to be closer to zero, so `mean` is no more than actual mean value
            # total_move = mean * len(self._bands) - num_all_subtasks
            #            <= actual_mean * len(self._bands) - num_all_subtasks = 0
            assert total_move <= 0
            if total_move != 0:
                band_move_nums[self._get_random_band(False)] -= total_move
            move_queued_subtasks.update(band_move_nums)
        return dict(sorted(move_queued_subtasks.items(), key=lambda item: item[1]))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from collections.abc import Iterable

import cloudpickle
import numpy as np
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy

from ..utils import estimate_pandas_size, pd_release_version

_HAS_SQUEEZE = pd_release_version < (1, 1, 0)


class GroupByWrapper:
    def __init__(
        self,
        obj,
        groupby_obj=None,
        keys=None,
        axis=0,
        level=None,
        grouper=None,
        exclusions=None,
        selection=None,
        as_index=True,
        sort=True,
        group_keys=True,
        squeeze=False,
        observed=False,
        mutated=False,
        grouper_cache=None,
    ):
        def fill_value(v, key):
            return (
                v if v is not None or groupby_obj is None else getattr(groupby_obj, key)
            )

        self.obj = obj
        self.keys = fill_value(keys, "keys")
        self.axis = fill_value(axis, "axis")
        self.level = fill_value(level, "level")
        self.exclusions = fill_value(exclusions, "exclusions")
        self.selection = selection
        self.as_index = fill_value(as_index, "as_index")
        self.sort = fill_value(sort, "sort")
        self.group_keys = fill_value(group_keys, "group_keys")
        self.squeeze = fill_value(squeeze, "squeeze")
        self.observed = fill_value(observed, "observed")
        self.mutated = fill_value(mutated, "mutated")

        if groupby_obj is None:
            groupby_kw = dict(
                keys=keys,
                axis=axis,
                level=level,
                grouper=grouper,
                exclusions=exclusions,
                as_index=as_index,
                group_keys=group_keys,
                squeeze=squeeze,
                observed=observed,
                mutated=mutated,
            )
            if not _HAS_SQUEEZE:  # pragma: no branch
                groupby_kw.pop("squeeze")

            if obj.ndim == 2:
                self.groupby_obj = DataFrameGroupBy(obj, **groupby_kw)
            else:
                self.groupby_obj = SeriesGroupBy(obj, **groupby_kw)
        else:
            self.groupby_obj = groupby_obj

        if grouper_cache:
            self.groupby_obj.grouper._cache = grouper_cache
        if selection:
            self.groupby_obj = self.groupby_obj[selection]

        self.is_frame = isinstance(self.groupby_obj, DataFrameGroupBy)

    def __getitem__(self, item):
        return GroupByWrapper(
            self.obj,
            keys=self.keys,
            axis=self.axis,
            level=self.level,
            grouper=self.groupby_obj.grouper,
            exclusions=self.exclusions,
            selection=item,
            as_index=self.as_index,
            sort=self.sort,
            group_keys=self.group_keys,
            squeeze=self.squeeze,
            observed=self.observed,
            mutated=self.mutated,
        )

    def __getattr__(self, item):
        if item.startswith("_"):  # pragma: no cover
            return object.__getattribute__(self, item)
        if item in getattr(self.obj, "columns", ()):
            return self.__getitem__(item)
        return getattr(self.groupby_obj, item)

    def __iter__(self):
        return self.groupby_obj.__iter__()

    def __sizeof__(self):
        return sys.getsizeof(self.obj) + sys.getsizeof(
            getattr(self.groupby_obj.grouper, "_cache", None)
        )

    def estimate_size(self):
        return estimate_pandas_size(self.obj) + estimate_pandas_size(self.obj.index)

    def __reduce__(self):
        return (
            type(self).from_tuple,
            (self.to_tuple(pickle_function=True, truncate=True),),
        )

    def __bool__(self):
        return bool(np.prod(self.shape))

    @property
    def empty(self):
        return self.obj.empty

    @property
    def shape(self):
        shape = list(self.groupby_obj.obj.shape)
        if self.is_frame and self.selection:
            shape[1] = len(self.selection)
        return tuple(shape)

    @property
    def _selected_obj(self):
        return getattr(self.groupby_obj, "_selected_obj")

    def to_tuple(self, truncate=False, pickle_function=False):
        if self.selection and truncate:
            if isinstance(self.selection, Iterable) and not isinstance(
                self.selection, str
            ):
                item_list = list(self.selection)
            else:
                item_list = [self.selection]
            item_set = set(item_list)

            if isinstance(self.keys, list):
                sel_keys = self.keys
            elif self.keys in self.obj.columns:
                sel_keys = [self.keys]
            else:
                sel_keys = []

            all_items = item_list + [k for k in sel_keys or () if k not in item_set]
            if set(all_items) == set(self.obj.columns):
                obj = self.obj
            else:
                obj = self.obj[all_items]
        else:
            obj = self.obj

        if pickle_function and callable(self.keys):
            keys = cloudpickle.dumps(self.keys)
        else:
            keys = self.keys

        return (
            obj,
            keys,
            self.axis,
            self.level,
            self.exclusions,
            self.selection,
            self.as_index,
            self.sort,
            self.group_keys,
            self.squeeze,
            self.observed,
            self.mutated,
            getattr(self.groupby_obj.grouper, "_cache", dict()),
        )

    @classmethod
    def from_tuple(cls, tp):
        (
            obj,
            keys,
            axis,
            level,
            exclusions,
            selection,
            as_index,
            sort,
            group_keys,
            squeeze,
            observed,
            mutated,
            grouper_cache,
        ) = tp

        if isinstance(keys, (bytes, bytearray)):
            keys = cloudpickle.loads(keys)

        return cls(
            obj,
            keys=keys,
            axis=axis,
            level=level,
            exclusions=exclusions,
            selection=selection,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            squeeze=squeeze,
            observed=observed,
            mutated=mutated,
            grouper_cache=grouper_cache,
        )


def wrapped_groupby(
    obj,
    by=None,
    axis=0,
    level=None,
    as_index=True,
    sort=True,
    group_keys=True,
    squeeze=False,
    observed=False,
):
    groupby_kw = dict(
        by=by,
        axis=axis,
        level=level,
        as_index=as_index,
        sort=sort,
        group_keys=group_keys,
        squeeze=squeeze,
        observed=observed,
    )
    if not _HAS_SQUEEZE:  # pragma: no branch
        groupby_kw.pop("squeeze")

    groupby_obj = obj.groupby(**groupby_kw)
    return GroupByWrapper(obj, groupby_obj=groupby_obj, as_index=as_index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import functools
import operator
import os
import warnings
import threading
from copy import deepcopy
from typing import Union, Dict


_DEFAULT_REDIRECT_WARN = (
    "Option {source} has been replaced by {target} and "
    "might be removed in a future release."
)


class OptionError(Exception):
    pass


class Redirection:
    def __init__(self, item, warn=None):
        self._items = item.split(".")
        self._warn = warn
        self._warned = True
        self._parent = None

    def bind(self, attr_dict):
        self._parent = attr_dict
        self.getvalue()
        self._warned = False

    def getvalue(self):
        if self._warn and not self._warned:
            self._warned = True
            warnings.warn(self._warn)
        conf = self._parent.root
        for it in self._items:
            conf = getattr(conf, it)
        return conf

    def setvalue(self, value):
        if self._warn and not self._warned:
            self._warned = True
            warnings.warn(self._warn)
        conf = self._parent.root
        for it in self._items[:-1]:
            conf = getattr(conf, it)
        setattr(conf, self._items[-1], value)


class AttributeDict(dict):
    def __init__(self, *args, **kwargs):
        self._inited = False
        self._parent = kwargs.pop("_parent", None)
        self._root = None
        super().__init__(*args, **kwargs)
        self._inited = True

    @property
    def root(self):
        if self._root is not None:
            return self._root
        if self._parent is None:
            self._root = self
        else:
            self._root = self._parent.root
        return self._root

    def __getattr__(self, item):
        if item in self:
            val = self[item]
            if isinstance(val, AttributeDict):
                return val
            elif isinstance(val[0], Redirection):
                return val[0].getvalue()
            else:
                return val[0]
        return object.__getattribute__(self, item)

    def __dir__(self):
        return list(self.keys())

    def register(self, key, value, validator=None):
        if isinstance(validator, tuple):
            validator = any_validator(*validator)
        self[key] = value, validator
        if isinstance(value, Redirection):
            value.bind(self)

    def unregister(self, key):
        del self[key]

    def _setattr(self, key, value, silent=False):
        splits = key.split(".")
        target = self
        for k in splits[:-1]:
            if not silent and (
                not isinstance(target, AttributeDict) or k not in target
            ):
                raise OptionError("You can only set the value of existing options")
            target = target[k]
        key = splits[-1]

        if not isinstance(value, AttributeDict):
            validate = None
            if key in target:
                val = target[key]
                validate = target[key][1]
                if validate is not None:
                    if not validate(value):
                        raise ValueError(f"Cannot set value `{value}`")
                if isinstance(val[0], Redirection):
                    val[0].setvalue(value)
                else:
                    target[key] = value, validate
            else:
                target[key] = value, validate
        else:
            target[key] = value

    def __setattr__(self, key, value):
        if key == "_inited":
            super().__setattr__(key, value)
            return
        try:
            object.__getattribute__(self, key)
            super().__setattr__(key, value)
            return
        except AttributeError:
            pass

        if not self._inited:
            super().__setattr__(key, value)
        else:
            self._setattr(key, value)

    def to_dict(self):
        d = dict()
        for k, v in self.items():
            if isinstance(v, AttributeDict):
                d.update(
                    {k + "." + sub_k: sub_v for sub_k, sub_v in v.to_dict().items()}
                )
            elif isinstance(v[0], Redirection):
                continue
            else:
                d[k] = v[0]
        return d


class Config:
    def __init__(self, config=None):
        self._config = config or AttributeDict()
        self._serialize_options = []

    def __dir__(self):
        return list(self._config.keys())

    def __getattr__(self, item):
        config = object.__getattribute__(self, "_config")
        return getattr(config, item)

    def __setattr__(self, key, value):
        if key.startswith("_"):
            object.__setattr__(self, key, value)
            return
        setattr(self._config, key, value)

    def register_option(self, option, value, validator=None, serialize=False):
        splits = option.split(".")
        conf = self._config
        if isinstance(validator, tuple):
            validator = any_validator(*validator)

        for name in splits[:-1]:
            config = conf.get(name)
            if config is None:
                val = AttributeDict(_parent=conf)
                conf[name] = val
                conf = val
            elif not isinstance(config, dict):
                raise AttributeError(
                    f"Fail to set option: {option}, conflict has encountered"
                )
            else:
                conf = config

        key = splits[-1]
        if conf.get(key) is not None:
            raise AttributeError(f"Fail to set option: {option}, option has been set")

        conf.register(key, value, validator)
        if serialize:
            self._serialize_options.append(option)

    def redirect_option(self, option, target, warn=_DEFAULT_REDIRECT_WARN):
        redir = Redirection(target, warn=warn.format(source=option, target=target))
        self.register_option(option, redir)

    def unregister_option(self, option):
        splits = option.split(".")
        conf = self._config
        for name in splits[:-1]:
            config = conf.get(name)
            if not isinstance(config, dict):
                raise AttributeError(
                    f"Fail to unregister option: {option}, conflict has encountered"
                )
            else:
                conf = config

        key = splits[-1]
        if key not in conf:
            raise AttributeError(
                f"Option {option} not configured, thus failed to unregister."
            )
        conf.unregister(key)

    def copy(self):
        new_options = Config(deepcopy(self._config))
        return new_options

    def update(self, new_config: Union["Config", Dict]):
        if not isinstance(new_config, dict):
            new_config = new_config._config
        for option, value in new_config.items():
            try:
                self.register_option(option, value)
            except AttributeError:
                setattr(self, option, value)

    def get_serializable(self):
        d = dict()
        for k in self._serialize_options:
            parts = k.split(".")
            v = self
            for p in parts:
                v = getattr(v, p)
            d[k] = v
        return d

    def fill_serialized(self, d):
        for k, v in d.items():
            parts = k.split(".")
            cf = self
            for p in parts[:-1]:
                cf = getattr(cf, p)
            setattr(cf, parts[-1], v)

    def to_dict(self):
        return self._config.to_dict()


@contextlib.contextmanager
def option_context(config=None):
    global_options = get_global_option()

    try:
        config = config or dict()
        local_options = Config(deepcopy(global_options._config))
        local_options.update(config)
        _options_local.default_options = local_options
        yield local_options
    finally:
        _options_local.default_options = global_options


def is_interactive():
    import __main__ as main

    return not hasattr(main, "__file__")


# validators
def any_validator(*validators):
    def validate(x):
        return any(validator(x) for validator in validators)

    return validate


def all_validator(*validators):
    def validate(x):
        return all(validator(x) for validator in validators)

    return validate


def _instance_check(typ, v):
    return isinstance(v, typ)


is_null = functools.partial(operator.is_, None)
is_bool = functools.partial(_instance_check, bool)
is_integer = functools.partial(_instance_check, int)
is_float = functools.partial(_instance_check, float)
is_numeric = functools.partial(_instance_check, (float, int))
is_string = functools.partial(_instance_check, str)
is_dict = functools.partial(_instance_check, dict)
is_list = functools.partial(_instance_check, list)


def is_in(vals):
    def validate(x):
        return x in vals

    return validate


default_options = Config()
default_options.register_option("tcp_timeout", 30, validator=is_integer)
default_options.register_option("verbose", False, validator=is_bool)
default_options.register_option("kv_store", ":inproc:", validator=is_string)
default_options.register_option("check_interval", 20, validator=is_integer)
default_options.register_option(
    "show_progress", "auto", validator=any_validator(is_bool, is_string)
)
default_options.register_option("serialize_method", "pickle")

# dataframe-related options
default_options.register_option(
    "dataframe.mode.use_inf_as_na", False, validator=is_bool
)
default_options.register_option(
    "dataframe.use_arrow_dtype", None, validator=any_validator(is_null, is_bool)
)
default_options.register_option(
    "dataframe.arrow_array.pandas_only", None, validator=any_validator(is_null, is_bool)
)

# learn options
assume_finite = os.environ.get("SKLEARN_ASSUME_FINITE")
if assume_finite is not None:
    assume_finite = bool(assume_finite)
working_memory = os.environ.get("SKLEARN_WORKING_MEMORY")
if working_memory is not None:
    working_memory = int(working_memory)
default_options.register_option(
    "learn.assume_finite", assume_finite, validator=any_validator(is_null, is_bool)
)
default_options.register_option(
    "learn.working_memory", working_memory, validator=any_validator(is_null, is_integer)
)

# the number of combined chunks in tree reduction or tree add
default_options.register_option("combine_size", 4, validator=is_integer, serialize=True)

# the default chunk store size
default_options.register_option(
    "chunk_store_limit", 128 * 1024**2, validator=is_numeric
)
default_options.register_option(
    "chunk_size", None, validator=any_validator(is_null, is_integer), serialize=True
)

# rechunk
default_options.register_option(
    "rechunk.threshold", 4, validator=is_integer, serialize=True
)
default_options.register_option(
    "rechunk.chunk_size_limit", int(1e8), validator=is_integer, serialize=True
)

default_options.register_option(
    "bincount.chunk_size_limit", int(1e8), validator=is_integer, serialize=True
)

# deploy
default_options.register_option("deploy.open_browser", True, validator=is_bool)

# optimization
default_options.register_option("optimize_tileable_graph", True, validator=is_bool)

# eager mode
default_options.register_option("eager_mode", False, validator=is_bool)

# optimization
default_options.register_option(
    "optimize.head_optimize_threshold", 1000, validator=is_integer
)

# debug
default_options.register_option("warn_duplicated_execution", False, validator=is_bool)

# client serialize type
default_options.register_option("client.serial_type", "arrow", validator=is_string)

# custom log dir
default_options.register_option(
    "custom_log_dir", None, validator=any_validator(is_null, is_string)
)

# vineyard
default_options.register_option(
    "vineyard.socket", os.environ.get("VINEYARD_IPC_SOCKET", None)
)
default_options.register_option(
    "vineyard.enabled", os.environ.get("WITH_VINEYARD", None) is not None
)

_options_local = threading.local()
_options_local.default_options = default_options


def get_global_option():
    ret = getattr(_options_local, "default_options", None)
    if ret is None:
        ret = _options_local.default_options = Config(deepcopy(default_options._config))

    return ret


class OptionsProxy:
    def __dir__(self):
        return dir(get_global_option())

    def __getattribute__(self, attr):
        return getattr(get_global_option(), attr)

    def __setattr__(self, key, value):
        setattr(get_global_option(), key, value)


options = OptionsProxy()

options.redirect_option("tensor.chunk_store_limit", "chunk_store_limit")
options.redirect_option("tensor.chunk_size", "chunk_size")
options.redirect_option("tensor.rechunk.threshold", "rechunk.threshold")
options.redirect_option("tensor.rechunk.chunk_size_limit", "rechunk.chunk_size_limit")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import Int32Field, StringField
from ...config import options
from ..utils import decide_chunk_sizes, get_order
from .diag import TensorDiagBase
from .core import TensorNoInput
from ...lib import sparse
from ..array_utils import create_array


class TensorEye(TensorNoInput, TensorDiagBase):
    _op_type_ = OperandDef.TENSOR_EYE

    _k = Int32Field("k")
    _order = StringField("order")

    def __init__(self, k=None, dtype=None, order=None, **kw):
        dtype = np.dtype(dtype or "f8")
        super().__init__(_k=k, dtype=dtype, _order=order, **kw)

    @property
    def k(self):
        return getattr(self, "_k", 0)

    @property
    def order(self):
        return self._order

    @classmethod
    def _get_nsplits(cls, op):
        tensor = op.outputs[0]
        chunk_size = tensor.extra_params.raw_chunk_size or options.chunk_size
        return decide_chunk_sizes(tensor.shape, chunk_size, tensor.dtype.itemsize)

    @classmethod
    def _get_chunk(cls, op, chunk_k, chunk_shape, chunk_idx):
        chunk_op = TensorEye(k=chunk_k, dtype=op.dtype, gpu=op.gpu, sparse=op.sparse)
        return chunk_op.new_chunk(None, shape=chunk_shape, index=chunk_idx)

    @classmethod
    def tile(cls, op):
        return (yield from TensorDiagBase.tile(op))

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        if op.sparse:
            ctx[chunk.key] = sparse.eye(
                chunk.shape[0], M=chunk.shape[1], k=op.k, dtype=op.dtype, gpu=op.gpu
            )
        else:
            ctx[chunk.key] = create_array(op)(
                "eye",
                chunk.shape[0],
                M=chunk.shape[1],
                k=op.k,
                dtype=op.dtype,
                order=op.order,
            )


def eye(N, M=None, k=0, dtype=None, sparse=False, gpu=None, chunk_size=None, order="C"):
    """
    Return a 2-D tensor with ones on the diagonal and zeros elsewhere.

    Parameters
    ----------
    N : int
      Number of rows in the output.
    M : int, optional
      Number of columns in the output. If None, defaults to `N`.
    k : int, optional
      Index of the diagonal: 0 (the default) refers to the main diagonal,
      a positive value refers to an upper diagonal, and a negative value
      to a lower diagonal.
    dtype : data-type, optional
      Data-type of the returned tensor.
    sparse: bool, optional
        Create sparse tensor if True, False as default
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    order : {'C', 'F'}, optional
        Whether the output should be stored in row-major (C-style) or
        column-major (Fortran-style) order in memory.

    Returns
    -------
    I : Tensor of shape (N,M)
      An tensor where all elements are equal to zero, except for the `k`-th
      diagonal, whose values are equal to one.

    See Also
    --------
    identity : (almost) equivalent function
    diag : diagonal 2-D tensor from a 1-D tensor specified by the user.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.eye(2, dtype=int).execute()
    array([[1, 0],
           [0, 1]])
    >>> mt.eye(3, k=1).execute()
    array([[ 0.,  1.,  0.],
           [ 0.,  0.,  1.],
           [ 0.,  0.,  0.]])

    """
    if M is None:
        M = N

    shape = (N, M)
    tensor_order = get_order(
        order,
        None,
        available_options="CF",
        err_msg="only 'C' or 'F' order is permitted",
    )
    op = TensorEye(k, dtype=dtype, gpu=gpu, sparse=sparse, order=order)
    return op(shape, chunk_size=chunk_size, order=tensor_order)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes
from ...serialization.serializables import (
    Int32Field,
    StringField,
    TupleField,
    FieldTypes,
)
from .base import Operand, VirtualOperand, OperandStage


class ShuffleProxy(VirtualOperand):
    _op_type_ = opcodes.SHUFFLE_PROXY


class MapReduceOperand(Operand):
    # for mapper
    mapper_id = Int32Field("mapper_id", default=0)
    # for reducer
    reducer_index = TupleField("reducer_index", FieldTypes.uint64)
    # Total reducer nums, which also be shuffle blocks for single mapper.
    n_reducers = Int32Field("n_reducers")
    # The reducer ordinal in all reducers. It's different from reducer_index,
    # which might be a tuple.
    reducer_ordinal = Int32Field("reducer_ordinal")
    reducer_phase = StringField("reducer_phase", default=None)

    def _new_chunks(self, inputs, kws=None, **kw):
        if getattr(self, "reducer_index", None) is None:
            index = None
            if kws:
                index = kws[0].get("index")
            self.reducer_index = index or kw.get("index")

        return super()._new_chunks(inputs, kws, **kw)

    def get_dependent_data_keys(self):
        from .fetch import FetchShuffle

        if self.stage == OperandStage.reduce:
            inputs = self.inputs or ()
            deps = []
            for inp in inputs:
                if isinstance(inp.op, ShuffleProxy):
                    deps.extend(
                        [(chunk.key, self.reducer_index) for chunk in inp.inputs or ()]
                    )
                elif isinstance(inp.op, FetchShuffle):
                    deps.extend([(k, self.reducer_index) for k in inp.op.source_keys])
                else:
                    deps.append(inp.key)
            return deps
        return super().get_dependent_data_keys()

    def _iter_mapper_key_idx_pairs(self, input_id=0, mapper_id=None):
        input_chunk = self.inputs[input_id]
        if isinstance(input_chunk.op, ShuffleProxy):
            keys = [inp.key for inp in input_chunk.inputs]
            idxes = [inp.index for inp in input_chunk.inputs]
            mappers = (
                [inp.op.mapper_id for inp in input_chunk.inputs]
                if mapper_id is not None
                else None
            )
        else:
            keys = input_chunk.op.source_keys
            idxes = input_chunk.op.source_idxes
            mappers = input_chunk.op.source_mappers

        if mapper_id is None:
            key_idx_pairs = zip(keys, idxes)
        else:
            key_idx_pairs = (
                (key, idx)
                for key, idx, mapper in zip(keys, idxes, mappers)
                if mapper == mapper_id
            )
        return key_idx_pairs

    def iter_mapper_data_with_index(
        self, ctx, input_id=0, mapper_id=None, pop=False, skip_none=False
    ):
        for key, idx in self._iter_mapper_key_idx_pairs(input_id, mapper_id):
            try:
                if pop:
                    yield idx, ctx.pop((key, self.reducer_index))
                else:
                    yield idx, ctx[key, self.reducer_index]
            except KeyError:
                if not skip_none:  # pragma: no cover
                    raise
                if not pop:
                    ctx[key, self.reducer_index] = None

    def iter_mapper_data(
        self, ctx, input_id=0, mapper_id=0, pop=False, skip_none=False
    ):
        for _idx, data in self.iter_mapper_data_with_index(
            ctx, input_id, mapper_id, pop=pop, skip_none=skip_none
        ):
            yield data



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import asyncio
import faulthandler
import glob
import importlib
import json
import logging.config
import os
import sys
import tempfile
from typing import List

import psutil

from ...utils import ensure_coverage
from ..utils import load_service_config_file, get_third_party_modules_from_config

logger = logging.getLogger(__name__)
_is_windows: bool = sys.platform.startswith("win")
ensure_coverage()


class OscarCommandRunner:
    command_description = None
    node_role = None
    _port_file_prefix = "mars_service_process"

    def __init__(self):
        faulthandler.enable()

        self.args = None
        self.ports = None
        self.config = {}
        self.pool = None

        self.logging_conf = {}

        self._running = False

    def config_args(self, parser):
        parser.add_argument("-e", "--endpoint", help="endpoint of the service")
        parser.add_argument("-H", "--host", help="host name of the service")
        parser.add_argument(
            "-p",
            "--ports",
            help="ports of the service, must equal to num of processes",
        )
        parser.add_argument("-c", "--config", help="service configuration")
        parser.add_argument(
            "-f", "--config-file", help="configuration file of the service"
        )
        parser.add_argument(
            "-s",
            "--supervisors",
            help="endpoint of supervisors, needed for workers and webs "
            "when kv-store argument is not available, or when you "
            "need to use multiple supervisors without kv-store",
        )
        parser.add_argument("--log-level", help="log level")
        parser.add_argument("--log-format", help="log format")
        parser.add_argument(
            "--log-conf", help="log config file, logging.conf by default"
        )
        parser.add_argument("--load-modules", nargs="*", help="modules to import")
        parser.add_argument(
            "--use-uvloop", help="use uvloop, 'auto' by default. Use 'no' to disable"
        )

    def _get_logging_config_paths(self):
        import mars

        log_conf = self.args.log_conf or "logging.conf"

        return [
            log_conf,
            os.path.join(os.path.abspath("."), log_conf),
            os.path.join(os.path.dirname(os.path.dirname(mars.__file__)), log_conf),
        ]

    def config_logging(self):
        for conf_path in self._get_logging_config_paths():
            if os.path.exists(conf_path):
                self.logging_conf["file"] = conf_path
                logging.config.fileConfig(conf_path, disable_existing_loggers=False)
                logger.debug("Use logging config file at %s", conf_path)
                break
        else:
            log_level = self.args.log_level
            level = getattr(logging, log_level.upper()) if log_level else logging.INFO
            logging.getLogger("__main__").setLevel(level)
            logging.getLogger("mars").setLevel(level)
            self.logging_conf["level"] = level
            if self.args.log_format:
                logging.basicConfig(format=self.args.log_format)
                self.logging_conf["format"] = self.args.log_format

    @classmethod
    def _build_endpoint_file_path(cls, pid: int = None, asterisk: bool = False):
        pid = pid or os.getpid()
        return os.path.join(
            tempfile.gettempdir(), f'{cls._port_file_prefix}.{"*" if asterisk else pid}'
        )

    def _write_supervisor_endpoint_file(self, args):
        file_name = self._build_endpoint_file_path()
        with open(file_name, "w") as port_file:
            port_file.write(args.endpoint)
        return file_name

    def _collect_supervisors_from_dir(self):
        endpoints = []
        for fn in glob.glob(self._build_endpoint_file_path(asterisk=True)):
            _, pid_str = os.path.basename(fn).rsplit(".", 1)
            # detect if process exists
            if pid_str.isdigit() and not psutil.pid_exists(int(pid_str)):
                continue
            with open(fn, "r") as ep_file:
                endpoints.append(ep_file.read().strip())
        return endpoints

    @classmethod
    def get_default_config_file(cls):
        mod_file_path = os.path.dirname(
            importlib.import_module(cls.__module__).__file__
        )
        return os.path.join(mod_file_path, "config.yml")

    def parse_args(self, parser, argv, environ=None):
        environ = environ or os.environ
        args = parser.parse_args(argv)

        if args.endpoint is not None and args.host is not None:  # pragma: no cover
            raise ValueError("Cannot specify host and endpoint at the same time")

        if "MARS_TASK_DETAIL" in environ:
            task_detail = json.loads(environ["MARS_TASK_DETAIL"])
            task_type, task_index = (
                task_detail["task"]["type"],
                task_detail["task"]["index"],
            )

            args.host = args.host or task_detail["cluster"][task_type][task_index]
            args.supervisors = args.supervisors or ",".join(
                task_detail["cluster"]["supervisor"]
            )

        default_host = "0.0.0.0" if not _is_windows else "127.0.0.1"
        env_host = os.environ.get(
            "MARS_BIND_HOST", os.environ.get("MARS_CONTAINER_IP", default_host)
        )
        args.host = args.host or env_host

        args.ports = args.ports or os.environ.get("MARS_BIND_PORT")
        if args.ports is not None:
            self.ports = [int(p) for p in args.ports.split(",")]

        if args.endpoint is None and len(self.ports or []) == 1:
            args.endpoint = f"{args.host}:{self.ports[0]}"
            self.ports = None

        args.use_uvloop = args.use_uvloop or "auto"

        if args.config is not None:
            self.config = json.loads(args.config)
        else:
            if args.config_file is None:
                args.config_file = self.get_default_config_file()
            self.config = load_service_config_file(args.config_file)

        load_modules = []
        for mods in list(args.load_modules or ()) + get_third_party_modules_from_config(
            self.config, self.node_role, environ
        ):
            load_modules.extend(mods.split(",") if mods else [])
        args.load_modules = tuple(load_modules)

        if args.supervisors is None:
            args.supervisors = ",".join(self._collect_supervisors_from_dir())

        return args

    async def _main(self, argv):
        self.config_logging()

        try:
            pool = self.pool = await self.create_actor_pool()

            await self.start_services()
            self._running = True
            await pool.join()
        except asyncio.CancelledError:
            if self._running:  # pragma: no branch
                await self.stop_services()
            if self.pool:  # pragma: no branch
                await self.pool.stop()

    async def create_actor_pool(self):
        raise NotImplementedError

    async def start_services(self):
        raise NotImplementedError

    async def stop_services(self):
        raise NotImplementedError

    def create_loop(self):
        use_uvloop = self.args.use_uvloop.strip()
        if use_uvloop in ("0", "no"):
            loop = asyncio.get_event_loop()
        else:
            try:
                import uvloop

                loop = uvloop.new_event_loop()
                asyncio.set_event_loop(loop)
            except ImportError:
                if use_uvloop == "auto":
                    loop = asyncio.get_event_loop()
                else:  # pragma: no cover
                    raise
        return loop

    def __call__(self, argv: List[str] = None):
        parser = argparse.ArgumentParser(description=self.command_description)
        self.config_args(parser)
        self.args = self.parse_args(parser, argv)

        loop = self.create_loop()
        task = loop.create_task(self._main(argv))

        try:
            loop.run_until_complete(task)
        except KeyboardInterrupt:
            task.cancel()
            loop.run_until_complete(task)
            # avoid displaying exception-unhandled warnings
            task.exception()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from functools import partial

try:
    from sklearn.exceptions import DataConversionWarning
except ImportError:  # pragma: no cover
    DataConversionWarning = None

from ....tensor.spatial import distance
from ...utils.validation import check_non_negative
from .core import PairwiseDistances
from .manhattan import manhattan_distances
from .cosine import cosine_distances
from .euclidean import euclidean_distances
from .haversine import haversine_distances


_VALID_METRICS = [
    "euclidean",
    "l2",
    "l1",
    "manhattan",
    "cityblock",
    "braycurtis",
    "canberra",
    "chebyshev",
    "correlation",
    "cosine",
    "dice",
    "hamming",
    "jaccard",
    "kulsinski",
    "mahalanobis",
    "matching",
    "minkowski",
    "rogerstanimoto",
    "russellrao",
    "seuclidean",
    "sokalmichener",
    "sokalsneath",
    "sqeuclidean",
    "yule",
    "wminkowski",
    "haversine",
]

# Helper functions - distance
PAIRWISE_DISTANCE_FUNCTIONS = {
    # If updating this dictionary, update the doc in both distance_metrics()
    # and also in pairwise_distances()!
    "cityblock": manhattan_distances,
    "cosine": cosine_distances,
    "euclidean": euclidean_distances,
    "haversine": haversine_distances,
    "l2": euclidean_distances,
    "l1": manhattan_distances,
    "manhattan": manhattan_distances,
    "precomputed": None,  # HACK: precomputed is always allowed, never called
}

# These distances recquire boolean tensors, when using mars.tensor.spatial.distance
PAIRWISE_BOOLEAN_FUNCTIONS = [
    "dice",
    "jaccard",
    "kulsinski",
    "matching",
    "rogerstanimoto",
    "russellrao",
    "sokalmichener",
    "sokalsneath",
    "yule",
]


def pairwise_distances(X, Y=None, metric="euclidean", **kwds):
    if (
        metric not in _VALID_METRICS
        and not callable(metric)
        and metric != "precomputed"
    ):
        raise ValueError(
            f"Unknown metric {metric}. Valid metrics are {_VALID_METRICS}, "
            "or 'precomputed', or a callable"
        )

    if metric == "precomputed":
        X, _ = PairwiseDistances.check_pairwise_arrays(X, Y, precomputed=True)

        whom = (
            "`pairwise_distances`. Precomputed distance "
            " need to have non-negative values."
        )
        X = check_non_negative(X, whom=whom)
        return X
    elif metric in PAIRWISE_DISTANCE_FUNCTIONS:
        func = PAIRWISE_DISTANCE_FUNCTIONS[metric]
    else:
        # including when metric is callable
        dtype = bool if metric in PAIRWISE_BOOLEAN_FUNCTIONS else None

        if (
            dtype == bool
            and (X.dtype != bool or (Y is not None and Y.dtype != bool))
            and DataConversionWarning is not None
        ):
            msg = f"Data was converted to boolean for metric {metric}"
            warnings.warn(msg, DataConversionWarning)

        X, Y = PairwiseDistances.check_pairwise_arrays(X, Y, dtype=dtype)
        if X is Y:
            return distance.squareform(distance.pdist(X, metric=metric, **kwds))
        func = partial(distance.cdist, metric=metric, **kwds)

    return func(X, Y, **kwds)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from dataclasses import dataclass
from typing import Tuple, Union, List

import numpy as np

from ...typing import ChunkType, TileableType
from ..utils import decide_chunk_sizes


chunk_size_type = Union[int, Tuple[int], Tuple[Tuple[int], ...]]


def get_nsplits(
    tileable: TileableType, new_chunk_size: chunk_size_type, itemsize: int
) -> Tuple[Tuple[int], ...]:
    if isinstance(new_chunk_size, dict):
        chunk_size = list(tileable.nsplits)
        for idx, c in new_chunk_size.items():
            chunk_size[idx] = c
    else:
        chunk_size = new_chunk_size

    return decide_chunk_sizes(tileable.shape, chunk_size, itemsize)


@dataclass
class RechunkInfo:
    out_index: Tuple[int]
    shape: Tuple[int]
    input_chunks: List[ChunkType]
    input_slices: List[Tuple[slice]]
    input_chunk_shape: List[int]


def gen_rechunk_infos(
    inp: TileableType, chunk_size: Tuple[Tuple[int], ...]
) -> List[RechunkInfo]:
    cum_in_nsplits = [np.cumsum(ns) for ns in inp.nsplits]
    cum_out_nsplits = [np.cumsum(ns) for ns in chunk_size]
    out_starts = [[0] + cum_ns[:-1].tolist() for cum_ns in cum_out_nsplits]
    out_ends = cum_out_nsplits
    out_start_indexes = [
        np.searchsorted(cum_ns, starts)
        for cum_ns, starts in zip(cum_in_nsplits, out_starts)
    ]
    out_end_indexes = [
        np.searchsorted(cum_ns, ends) for cum_ns, ends in zip(cum_in_nsplits, out_ends)
    ]

    chunk_index_iter = itertools.product(*(range(len(s)) for s in chunk_size))
    rechunk_infos = []
    for chunk_index in chunk_index_iter:
        shape = tuple(chunk_size[dim][i] for dim, i in enumerate(chunk_index))
        inp_chunk_slices = [list() for _ in range(len(chunk_index))]
        inp_chunk_indexes = [list() for _ in range(len(chunk_index))]
        for dim, i in enumerate(chunk_index):
            size_start = out_starts[dim][i]
            size_end = out_ends[dim][i]
            start_index = out_start_indexes[dim][i]
            end_index = out_end_indexes[dim][i]
            for inp_i in range(start_index, end_index + 1):
                inp_start = cum_in_nsplits[dim][inp_i - 1] if inp_i > 0 else 0
                inp_end = cum_in_nsplits[dim][inp_i]
                slice_start = max(inp_start, size_start) - inp_start
                slice_end = min(inp_end, size_end) - inp_start
                if slice_start == 0 and slice_end == inp_end - inp_start:
                    # slice all
                    slc = slice(None)
                elif slice_start == slice_end and size_start != size_end:
                    continue
                else:
                    slc = slice(slice_start, slice_end)
                inp_chunk_slices[dim].append(slc)
                inp_chunk_indexes[dim].append(inp_i)

        inp_chunks = []
        inp_slices = []
        rechunk_info = RechunkInfo(
            out_index=chunk_index,
            shape=shape,
            input_chunks=inp_chunks,
            input_slices=inp_slices,
            input_chunk_shape=list(len(s) for s in inp_chunk_indexes),
        )
        for inp_chunk_index, inp_chunk_slice in zip(
            itertools.product(*inp_chunk_indexes),
            itertools.product(*inp_chunk_slices),
        ):
            inp_chunk = inp.cix[tuple(inp_chunk_index)]
            inp_chunks.append(inp_chunk)
            inp_slices.append(inp_chunk_slice)
        rechunk_infos.append(rechunk_info)

    return rechunk_infos



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import copy
from typing import Dict, Union

from ...oscar.profiling import ProfilingData
from ...utils import Timer
from ..errors import ServerClosed
from .communication import Client
from .message import _MessageBase, ResultMessage, ErrorMessage, DeserializeMessageFailed
from .router import Router


ResultMessageType = Union[ResultMessage, ErrorMessage]


class ActorCaller:
    __slots__ = "_client_to_message_futures", "_clients"

    def __init__(self):
        self._client_to_message_futures: Dict[
            Client, Dict[bytes, asyncio.Future]
        ] = dict()
        self._clients: Dict[Client, asyncio.Task] = dict()

    async def get_client(self, router: Router, dest_address: str) -> Client:
        client = await router.get_client(dest_address, from_who=self)
        if client not in self._clients:
            self._clients[client] = asyncio.create_task(self._listen(client))
            self._client_to_message_futures[client] = dict()
        return client

    async def _listen(self, client: Client):
        while not client.closed:
            try:
                try:
                    message: _MessageBase = await client.recv()
                except (EOFError, ConnectionError, BrokenPipeError):
                    # remote server closed, close client and raise ServerClosed
                    try:
                        await client.close()
                    except (ConnectionError, BrokenPipeError):
                        # close failed, ignore it
                        pass
                    raise ServerClosed(
                        f"Remote server {client.dest_address} closed"
                    ) from None
                future = self._client_to_message_futures[client].pop(message.message_id)
                future.set_result(message)
            except DeserializeMessageFailed as e:
                message_id = e.message_id
                future = self._client_to_message_futures[client].pop(message_id)
                future.set_exception(e.__cause__)
            except Exception as e:  # noqa: E722  # pylint: disable=bare-except
                message_futures = self._client_to_message_futures.get(client)
                self._client_to_message_futures[client] = dict()
                for future in message_futures.values():
                    future.set_exception(copy.copy(e))
            finally:
                await asyncio.sleep(0)

        message_futures = self._client_to_message_futures.get(client)
        self._client_to_message_futures[client] = dict()
        error = ServerClosed(f"Remote server {client.dest_address} closed")
        for future in message_futures.values():
            future.set_exception(copy.copy(error))

    async def call(
        self,
        router: Router,
        dest_address: str,
        message: _MessageBase,
        wait: bool = True,
    ) -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        client = await self.get_client(router, dest_address)
        loop = asyncio.get_running_loop()
        wait_response = loop.create_future()
        self._client_to_message_futures[client][message.message_id] = wait_response

        with Timer() as timer:
            try:
                await client.send(message)
            except ConnectionError:
                try:
                    await client.close()
                except ConnectionError:
                    # close failed, ignore it
                    pass
                raise ServerClosed(f"Remote server {client.dest_address} closed")

            if not wait:
                r = wait_response
            else:
                r = await wait_response

        ProfilingData.collect_actor_call(message, timer.duration)
        return r

    async def stop(self):
        try:
            await asyncio.gather(*[client.close() for client in self._clients])
        except (ConnectionError, ServerClosed):
            pass
        self.cancel_tasks()

    def cancel_tasks(self):
        # cancel listening for all clients
        _ = [task.cancel() for task in self._clients.values()]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...core import ENTITY_TYPE, CHUNK_TYPE, recursive_tile
from ...core.custom_log import redirect_custom_log
from ...serialization.serializables import (
    FunctionField,
    BoolField,
    TupleField,
    DictField,
)
from ...utils import (
    enter_current_session,
    quiet_stdio,
    find_objects,
    replace_objects,
    has_unknown_shape,
)
from ..operands import TensorOperand, TensorOperandMixin


class TensorMapChunk(TensorOperand, TensorOperandMixin):
    _op_type_ = opcodes.MAP_CHUNK

    _func = FunctionField("func")
    _elementwise = BoolField("elementwise")
    _args = TupleField("args")
    _kwargs = DictField("kwargs")
    _with_chunk_index = BoolField("with_chunk_index")

    def __init__(
        self,
        func=None,
        args=None,
        kwargs=None,
        elementwise=None,
        with_chunk_index=None,
        **kw
    ):
        super().__init__(
            _func=func,
            _args=args,
            _kwargs=kwargs,
            _elementwise=elementwise,
            _with_chunk_index=with_chunk_index,
            **kw
        )

    @property
    def func(self):
        return self._func

    @property
    def elementwise(self):
        return self._elementwise

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def with_chunk_index(self):
        return self._with_chunk_index

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        old_inputs = find_objects(self._args, ENTITY_TYPE) + find_objects(
            self._kwargs, ENTITY_TYPE
        )
        mapping = {o: n for o, n in zip(old_inputs, self._inputs[1:])}
        self._args = replace_objects(self._args, mapping)
        self._kwargs = replace_objects(self._kwargs, mapping)

    def __call__(self, t, dtype=None, shape=None):
        if dtype is None:
            try:
                kwargs = self.kwargs or dict()
                if self.with_chunk_index:
                    kwargs["chunk_index"] = (0,) * t.ndim
                with np.errstate(all="ignore"), quiet_stdio():
                    mock_result = self.func(
                        np.random.rand(2, 2).astype(t.dtype),
                        *(self.args or ()),
                        **kwargs
                    )
            except:
                raise TypeError("Cannot estimate output type of map_chunk call")
            dtype = mock_result.dtype

        if shape is not None:
            new_shape = shape
        else:
            new_shape = t.shape if self.elementwise else (np.nan,) * t.ndim
        inputs = (
            [t]
            + find_objects(self.args, ENTITY_TYPE)
            + find_objects(self.kwargs, ENTITY_TYPE)
        )
        return self.new_tensor(inputs, dtype=dtype, shape=new_shape)

    @classmethod
    def tile(cls, op: "TensorMapChunk"):
        inp = op.inputs[0]
        out = op.outputs[0]

        new_inputs = [op.inputs[0]]
        if has_unknown_shape(*op.inputs[1:]):
            yield
        for other_inp in op.inputs[1:]:
            other_inp = yield from recursive_tile(other_inp.rechunk(other_inp.shape))
            new_inputs.append(other_inp)

        chunks = []
        for c in inp.chunks:
            params = c.params
            params["dtype"] = out.dtype
            if not op.elementwise:
                params["shape"] = (np.nan,) * out.ndim
                params["index"] = params["index"][: out.ndim]

            new_op = op.copy().reset_key()
            new_op.tileable_op_key = out.key
            chunk_inputs = [c]
            for other_inp in new_inputs[1:]:
                chunk_inputs.append(other_inp.chunks[0])
            chunks.append(new_op.new_chunk(chunk_inputs, **params))

        new_op = op.copy().reset_key()
        params = out.params
        nsplits = inp.nsplits[: out.ndim]
        if not op.elementwise:
            nsplits = tuple((np.nan,) * len(sp) for sp in nsplits)
        return new_op.new_tileables(op.inputs, chunks=chunks, nsplits=nsplits, **params)

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: "TensorMapChunk"):
        in_data = ctx[op.inputs[0].key]
        out_chunk = op.outputs[0]

        args = op.args or tuple()
        kwargs = op.kwargs or dict()
        if op.with_chunk_index:
            kwargs["chunk_index"] = out_chunk.index

        chunks = find_objects(args, CHUNK_TYPE) + find_objects(kwargs, CHUNK_TYPE)
        mapping = {chunk: ctx[chunk.key] for chunk in chunks}
        args = replace_objects(args, mapping)
        kwargs = replace_objects(kwargs, mapping)

        ctx[op.outputs[0].key] = op.func(in_data, *args, **kwargs)


def map_chunk(t, func, args=(), **kwargs):
    """
    Apply function to each chunk.

    Parameters
    ----------
    func : function
        Function to apply to each chunk.
    args : tuple
        Positional arguments to pass to func in addition to the array.
    **kwargs
        Additional keyword arguments to pass as keywords arguments to func.

    Returns
    -------
    Tensor
        Result of applying ``func`` to each chunk of the Tensor.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.array([[4, 9]] * 3)
    >>> a.execute()
    array([[4, 9],
           [4, 9],
           [4, 9]])

    Output dtype will be auto inferred.

    >>> a.map_chunk(lambda c: c * 0.5).execute()
    array([[2. , 4.5],
           [2. , 4.5],
           [2. , 4.5]])

    You can specify ``dtype`` by yourself if auto infer failed.
    """
    elementwise = kwargs.pop("elementwise", None)
    dtype = np.dtype(kwargs.pop("dtype")) if "dtype" in kwargs else None
    shape = kwargs.pop("shape", None)
    with_chunk_index = kwargs.pop("with_chunk_index", False)

    op = TensorMapChunk(
        func=func,
        args=args,
        kwargs=kwargs,
        elementwise=elementwise,
        with_chunk_index=with_chunk_index,
    )
    return op(t, dtype=dtype, shape=shape)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .average import average
from .bincount import bincount
from .corrcoef import corrcoef
from .cov import cov
from .digitize import digitize, TensorDigitize
from .ptp import ptp
from .histogram import (
    histogram_bin_edges,
    TensorHistogramBinEdges,
    histogram,
    TensorHistogram,
)
from .quantile import quantile
from .percentile import percentile
from .median import median


def _install():
    from ..core import Tensor, TensorData

    for cls in (Tensor, TensorData):
        setattr(cls, "ptp", ptp)


_install()
del _install



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pandas as pd
import numpy as np

from ... import opcodes as OperandDef
from ...core import OutputType
from ...tensor.core import TensorOrder
from ...utils import lazy_import
from .core import DataFrameReductionOperand, DataFrameReductionMixin, CustomReduction

cudf = lazy_import("cudf")


class UniqueReduction(CustomReduction):
    def agg(self, data):  # noqa: W0221  # pylint: disable=arguments-differ
        xdf = cudf if self.is_gpu() else pd
        # convert to series data
        return xdf.Series(data.unique())

    def post(self, data):  # noqa: W0221  # pylint: disable=arguments-differ
        return data.unique()


class DataFrameUnique(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.UNIQUE
    _func_name = "unique"

    @classmethod
    def get_reduction_callable(cls, op):
        return UniqueReduction(name=cls._func_name, is_gpu=op.is_gpu())

    @classmethod
    def tile(cls, op):
        if op.method == "tree":
            return (yield from super().tile(op))
        else:
            raise NotImplementedError(f"Method {op.method} hasn't been supported")

    def __call__(self, a):
        self.output_types = [OutputType.tensor]
        return self.new_tileables(
            [a], shape=(np.nan,), dtype=a.dtype, order=TensorOrder.C_ORDER
        )[0]


def unique(values, method="tree"):
    """
    Uniques are returned in order of appearance. This does NOT sort.

    Parameters
    ----------
    values : 1d array-like
    method : 'shuffle' or 'tree', 'tree' method provide a better performance, 'shuffle'
    is recommended if the number of unique values is very large.
    See Also
    --------
    Index.unique
    Series.unique

    Examples
    --------
    >>> import mars.dataframe as md
    >>> import pandas as pd
    >>> md.unique(md.Series([2, 1, 3, 3])).execute()
    array([2, 1, 3])

    >>> md.unique(md.Series([2] + [1] * 5)).execute()
    array([2, 1])

    >>> md.unique(md.Series([pd.Timestamp('20160101'),
    ...                     pd.Timestamp('20160101')])).execute()
    array(['2016-01-01T00:00:00.000000000'], dtype='datetime64[ns]')

    >>> md.unique(md.Series([pd.Timestamp('20160101', tz='US/Eastern'),
    ...                      pd.Timestamp('20160101', tz='US/Eastern')])).execute()
    array([Timestamp('2016-01-01 00:00:00-0500', tz='US/Eastern')],
          dtype=object)
    """
    op = DataFrameUnique(method=method)
    return op(values)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Dict

from .... import oscar as mo
from ...session import SessionAPI
from ..store import get_meta_store


class MetaStoreManagerActor(mo.Actor):
    def __init__(self, meta_store_name: str, config: Dict):
        self._meta_store_name = meta_store_name
        self._meta_store_type = get_meta_store(meta_store_name)
        self._config = config
        self._meta_init_kwargs = None

        # API
        self._session_api = None

    async def __post_create__(self):
        self._meta_init_kwargs = await self._meta_store_type.create(self._config)
        self._session_api = await SessionAPI.create(self.address)

    async def new_session_meta_store(self, session_id: str) -> mo.ActorRef:
        session_address = await self._session_api.get_session_address(session_id)
        allocate_strategy = mo.allocate_strategy.AddressSpecified(session_address)
        return await mo.create_actor(
            MetaStoreActor,
            self._meta_store_name,
            session_id,
            address=self.address,
            uid=MetaStoreActor.gen_uid(session_id),
            allocate_strategy=allocate_strategy,
            **self._meta_init_kwargs,
        )


class MetaStoreActor(mo.Actor):
    def __init__(self, meta_store_name: str, session_id: str, **meta_store_kwargs):
        meta_store_type = get_meta_store(meta_store_name)
        self._store = meta_store_type(session_id, **meta_store_kwargs)
        self._worker_meta_store_refs = []

    def add_worker_meta_store(self, ref: mo.ActorRef):
        self._worker_meta_store_refs.append(ref)

    async def __pre_destroy__(self):
        await asyncio.gather(
            *[
                mo.destroy_actor(mo.create_actor_ref(ref))
                for ref in self._worker_meta_store_refs
            ]
        )

    @staticmethod
    def gen_uid(session_id: str):
        return f"{session_id}_meta"

    def __getattr__(self, attr):
        return getattr(self._store, attr)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    import tiledb
except (ImportError, OSError):  # pragma: no cover
    tildb = None

from ... import opcodes as OperandDef
from ...serialization.serializables import (
    FieldTypes,
    DictField,
    TupleField,
    StringField,
    Int64Field,
    KeyField,
)
from ...lib.sparse import SparseNDArray
from ...lib.sparse.core import sps
from ..datasource import tensor as astensor
from ..operands import TensorOperandMixin, TensorOperand
from ..utils import get_tiledb_ctx
from .core import TensorDataStore
from .utils import check_tiledb_array_with_tensor, get_tiledb_schema_from_tensor


class TensorTileDBDataStore(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_TILEDB

    _input = KeyField("input")
    _tiledb_config = DictField("tiledb_config")
    # URI of array to write
    _tiledb_uri = StringField("tiledb_uri")
    # encryption key to decrypt if provided
    _tiledb_key = StringField("tiledb_key")
    # open array at a given timestamp if provided
    _tiledb_timestamp = Int64Field("tiledb_timestamp")
    _axis_offsets = TupleField("axis_offsets", FieldTypes.int64)

    def __init__(
        self,
        tiledb_config=None,
        tiledb_uri=None,
        tiledb_key=None,
        tiledb_timestamp=None,
        **kw,
    ):
        super().__init__(
            _tiledb_config=tiledb_config,
            _tiledb_uri=tiledb_uri,
            _tiledb_key=tiledb_key,
            _tiledb_timestamp=tiledb_timestamp,
            **kw,
        )

    @property
    def tiledb_config(self):
        return self._tiledb_config

    @property
    def tiledb_uri(self):
        return self._tiledb_uri

    @property
    def tiledb_key(self):
        return self._tiledb_key

    @property
    def tiledb_timestamp(self):
        return self._tiledb_timestamp

    @property
    def axis_offsets(self):
        return self._axis_offsets

    @classmethod
    def _get_out_chunk(cls, op, in_chunk):
        chunk_op = op.copy().reset_key()
        nsplits = op.input.nsplits
        axis_offsets = []
        for axis, idx in enumerate(in_chunk.index):
            axis_offsets.append(sum(nsplits[axis][:idx]))
        chunk_op._axis_offsets = tuple(axis_offsets)
        out_chunk_shape = (0,) * in_chunk.ndim
        return chunk_op.new_chunk(
            [in_chunk], shape=out_chunk_shape, index=in_chunk.index
        )

    @classmethod
    def _process_out_chunks(cls, op, out_chunks):
        if len(out_chunks) == 1:
            return out_chunks

        consolidate_op = TensorTileDBConsolidate(
            tiledb_config=op.tiledb_config,
            tiledb_uri=op.tiledb_uri,
            tiledb_key=op.tiledb_key,
            sparse=op.sparse,
            dtype=op.dtype,
        )
        return consolidate_op.new_chunks(
            out_chunks, shape=out_chunks[0].shape, index=(0,) * out_chunks[0].ndim
        )

    @classmethod
    def tile(cls, op):
        import tiledb

        tensor = super().tile(op)[0]

        ctx = tiledb.Ctx(op.tiledb_config)
        tiledb_array_type = (
            tiledb.SparseArray if tensor.issparse() else tiledb.DenseArray
        )
        try:
            tiledb_array_type(
                uri=op.tiledb_uri,
                key=op.tiledb_key,
                timestamp=op.tiledb_timestamp,
                ctx=ctx,
            )
        except tiledb.TileDBError:
            # not exist, try to create TileDB Array by given uri
            tiledb_array_schema = get_tiledb_schema_from_tensor(
                op.input, ctx, op.input.nsplits
            )
            tiledb_array_type.create(
                op.tiledb_uri, tiledb_array_schema, key=op.tiledb_key
            )

        return [tensor]

    @classmethod
    def execute(cls, ctx, op):
        tiledb_ctx = get_tiledb_ctx(op.tiledb_config)
        uri = op.tiledb_uri
        key = op.tiledb_key
        timestamp = op.tiledb_timestamp
        axis_offsets = op.axis_offsets

        chunk = op.outputs[0]
        if not chunk.issparse():
            # dense
            to_store = np.ascontiguousarray(ctx[op.input.key])
            slcs = []
            for axis in range(chunk.ndim):
                axis_offset = int(axis_offsets[axis])
                axis_length = int(op.input.shape[axis])
                slcs.append(slice(axis_offset, axis_offset + axis_length))
            with tiledb.DenseArray(
                uri=uri, ctx=tiledb_ctx, mode="w", key=key, timestamp=timestamp
            ) as arr:
                arr[tuple(slcs)] = to_store
            ctx[chunk.key] = np.empty((0,) * chunk.ndim, dtype=chunk.dtype)
        else:
            # sparse
            to_store = ctx[op.input.key].spmatrix.tocoo()
            if to_store.nnz > 0:
                with tiledb.SparseArray(
                    uri=uri, ctx=tiledb_ctx, mode="w", key=key, timestamp=timestamp
                ) as arr:
                    if chunk.ndim == 1:
                        vec = to_store.col if to_store.shape[0] == 1 else to_store.row
                        vec += axis_offsets[0]
                        arr[vec] = to_store.data
                    else:
                        i, j = (
                            to_store.row + axis_offsets[0],
                            to_store.col + axis_offsets[1],
                        )
                        arr[i, j] = to_store.data
            ctx[chunk.key] = SparseNDArray(
                sps.csr_matrix((0, 0), dtype=chunk.dtype), shape=chunk.shape
            )


class TensorTileDBConsolidate(TensorOperandMixin, TensorOperand):
    _op_type_ = OperandDef.TENSOR_STORE_TILEDB_CONSOLIDATE

    _tiledb_config = DictField("tiledb_config")
    # URI of array to write
    _tiledb_uri = StringField("tiledb_uri")
    # encryption key to decrypt if provided
    _tiledb_key = StringField("tiledb_key")

    def __init__(self, tiledb_config=None, tiledb_uri=None, tiledb_key=None, **kw):
        super().__init__(
            _tiledb_config=tiledb_config,
            _tiledb_uri=tiledb_uri,
            _tiledb_key=tiledb_key,
            **kw,
        )

    def calc_shape(self, *inputs_shape):
        return self.outputs[0].shape

    @property
    def tiledb_config(self):
        return self._tiledb_config

    @property
    def tiledb_uri(self):
        return self._tiledb_uri

    @property
    def tiledb_key(self):
        return self._tiledb_key

    @classmethod
    def tile(cls, op):
        raise TypeError(f"{cls.__name__} is a chunk op, cannot be tiled")

    @classmethod
    def execute(cls, ctx, op):
        tiledb_config = tiledb.Config(op.tiledb_config)
        uri = op.tiledb_uri
        key = op.tiledb_key

        tiledb.consolidate(config=tiledb_config, uri=uri, key=key)
        ctx[op.outputs[0].key] = ctx[op.inputs[0].key]


def totiledb(uri, x, ctx=None, key=None, timestamp=None):
    import tiledb

    x = astensor(x)
    raw_ctx = ctx
    if raw_ctx is None:
        ctx = tiledb.Ctx()

    tiledb_array_type = tiledb.SparseArray if x.issparse() else tiledb.DenseArray
    try:
        tiledb_array = tiledb_array_type(uri=uri, key=key, timestamp=timestamp, ctx=ctx)
        # if already created, we will check the shape and dtype
        check_tiledb_array_with_tensor(x, tiledb_array)
    except tiledb.TileDBError:
        # not exist, as we don't know the tile,
        # we will create the tiledb array in the tile of tensor
        pass

    tiledb_config = None if raw_ctx is None else raw_ctx.config().dict()
    op = TensorTileDBDataStore(
        tiledb_config=tiledb_config,
        tiledb_uri=uri,
        tiledb_key=key,
        tiledb_timestamp=timestamp,
        dtype=x.dtype,
        sparse=x.issparse(),
    )
    return op(x)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import enum
import logging
import os
import posixpath
from urllib.parse import urlparse, unquote

from ....utils import lazy_import, lazy_import_on_load

ray = lazy_import("ray")

logger = logging.getLogger(__name__)


def get_placement_group(pg_name):  # pragma: no cover
    return ray.util.get_placement_group(pg_name)


def process_address_to_placement(address):
    """
    Parameters
    ----------
    address: str
        The address of an actor pool which running in a ray actor. It's also
        the name of the ray actor. address ex: ray://${pg_name}/${bundle_index}/${process_index}

    Returns
    -------
    tuple
        A tuple consisting of placement group name, bundle index, process index.
    """
    name, parts = _address_to_placement(address)
    if not parts or len(parts) != 2:
        raise ValueError(
            f"Only bundle index and process index path are allowed in ray "
            f"address {address} but got {parts}."
        )
    bundle_index, process_index = parts
    return name, int(bundle_index), int(process_index)


def node_address_to_placement(address):
    """
    Parameters
    ----------
    address : str
        The address of a node. ex: ray://${pg_name}/${bundle_index}

    Returns
    -------
    tuple
        A tuple consisting of placement group name, bundle index.
    """
    name, parts = _address_to_placement(address)
    if not parts or len(parts) != 1:
        raise ValueError(
            f"Only bundle index path is allowed in ray address {address} but got {parts}"
        )
    bundle_index = parts[0]
    return name, int(bundle_index)


def _address_to_placement(address):
    """

    Parameters
    ----------
    address : str
        The address of a node or an actor pool which running in a ray actor.

    Returns
    -------
    tuple
        A tuple consisting of placement group name, bundle index, process index.
    """
    parsed_url = urlparse(unquote(address))
    if parsed_url.scheme != "ray":
        raise ValueError(f"The address scheme is not ray: {address}")
    # os.path.split will not handle backslashes (\) correctly,
    # so we use the posixpath.
    parts = []
    if parsed_url.netloc:
        tmp = parsed_url.path
        while tmp and tmp != "/":
            tmp2, item = posixpath.split(tmp)
            parts.append(item)
            if tmp2 != tmp:
                tmp = tmp2
            else:
                parts.append(tmp2)
                break
    parts = list(reversed(parts))
    return parsed_url.netloc, parts


def process_placement_to_address(
    pg_name: str, bundle_index: int, process_index: int = 0
):
    return f"ray://{pg_name}/{bundle_index}/{process_index}"


def node_placement_to_address(pg_name, bundle_index):
    return f"ray://{pg_name}/{bundle_index}"


def addresses_to_placement_group_info(address_to_resources):
    bundles = {}
    pg_name = None
    for address, bundle_resources in address_to_resources.items():
        name, bundle_index = node_address_to_placement(address)
        if pg_name is None:
            pg_name = name
        else:
            if name != pg_name:
                raise ValueError(
                    "All addresses should have consistent placement group names."
                )
        bundles[bundle_index] = bundle_resources
    sorted_bundle_keys = sorted(bundles.keys())
    if sorted_bundle_keys != list(range(len(address_to_resources))):
        raise ValueError("The addresses contains invalid bundle.")
    bundles = [bundles[k] for k in sorted_bundle_keys]
    if not pg_name:
        raise ValueError("Can't find a valid placement group name.")
    return pg_name, bundles


def placement_group_info_to_addresses(pg_name, bundles):
    addresses = {}
    for bundle_index, bundle_resources in enumerate(bundles):
        address = node_placement_to_address(pg_name, bundle_index)
        addresses[address] = bundle_resources
    return addresses


async def kill_and_wait(
    actor_handle: "ray.actor.ActorHandle", no_restart=False, timeout: float = 30
):
    if "COV_CORE_SOURCE" in os.environ:  # pragma: no cover
        try:
            # must clean up first, or coverage info lost
            await actor_handle.cleanup.remote()
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            pass
    r = actor_handle.wait.remote(timeout)
    ray.kill(actor_handle, no_restart=no_restart)
    ready, _ = await asyncio.wait([r], timeout=timeout)
    if ready:
        try:
            await r
        except ray.exceptions.RayActorError:
            return  # We expect a RayActorError, it indicated that the actor is died.
    raise Exception(
        f"The actor {actor_handle} is not died after ray.kill {timeout} seconds."
    )


@lazy_import_on_load(ray)
def _patch_event_security():
    global ray

    if ray and not hasattr(ray, "report_event"):  # pragma: no cover
        # lower version of ray doesn't support event

        class EventSeverity(enum.Enum):
            INFO = 0
            WARNING = 1
            ERROR = 2
            FATAL = 3

        def _report_event(severity, label, message):
            logger.warning(
                "severity: %s, label: %s, message: %s.", severity, label, message
            )

        import ray

        ray.EventSeverity = EventSeverity
        ray.report_event = _report_event


def report_event(severity, label, message):
    if ray and ray.is_initialized():
        severity = (
            getattr(ray.EventSeverity, severity)
            if isinstance(severity, str)
            else severity
        )
        ray.report_event(severity, label, message)



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ..oscar.worker import WorkerCommandRunner
from .core import K8SServiceMixin

logger = logging.getLogger(__name__)


class K8SWorkerCommandRunner(K8SServiceMixin, WorkerCommandRunner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def start_services(self):
        from ..oscar.worker import start_worker
        from ...services.cluster import ClusterAPI

        self.write_pid_file()
        await start_worker(
            self.pool.external_address,
            self.args.supervisors,
            self.band_to_resource,
            list(self.args.load_modules),
            self.config,
            mark_ready=False,
        )
        await self.wait_all_supervisors_ready()

        cluster_api = await ClusterAPI.create(self.args.endpoint)
        await cluster_api.mark_node_ready()

        await self.start_readiness_server()

    async def stop_services(self):
        await self.stop_readiness_server()
        await super().stop_services()


main = K8SWorkerCommandRunner()

if __name__ == "__main__":  # pragma: no branch
    main()
